# Copyright 2020 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

from neutron_lib import constants
from neutron_lib.services.qos import constants as qos_constants

from neutron.common.ovn import utils as ovn_utils
from neutron.db import l3_db
from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb.extensions \
    import qos as qos_extension
from neutron.tests.functional import base


QOS_RULE_BW_1 = {'max_kbps': 200, 'max_burst_kbps': 100}
QOS_RULE_BW_2 = {'max_kbps': 300}
QOS_RULE_DSCP_1 = {'dscp_mark': 16}
QOS_RULE_DSCP_2 = {'dscp_mark': 20}

QOS_RULES_1 = {
    constants.EGRESS_DIRECTION: {
        qos_constants.RULE_TYPE_BANDWIDTH_LIMIT: QOS_RULE_BW_1,
        qos_constants.RULE_TYPE_DSCP_MARKING: QOS_RULE_DSCP_1},
    constants.INGRESS_DIRECTION: {
        qos_constants.RULE_TYPE_BANDWIDTH_LIMIT: QOS_RULE_BW_2}
}

QOS_RULES_2 = {
    constants.EGRESS_DIRECTION: {
        qos_constants.RULE_TYPE_BANDWIDTH_LIMIT: QOS_RULE_BW_2,
        qos_constants.RULE_TYPE_DSCP_MARKING: QOS_RULE_DSCP_2}
}

QOS_RULES_3 = {
    constants.INGRESS_DIRECTION: {
        qos_constants.RULE_TYPE_BANDWIDTH_LIMIT: QOS_RULE_BW_1}
}


class TestOVNClientQosExtension(base.TestOVNFunctionalBase):

    def setUp(self, maintenance_worker=False):
        super(TestOVNClientQosExtension, self).setUp(
            maintenance_worker=maintenance_worker)
        self._add_logical_switch()
        self.qos_driver = qos_extension.OVNClientQosExtension(
            nb_idl=self.nb_api)
        self.gw_port_id = 'gw_port_id'
        self._mock_get_router = mock.patch.object(l3_db.L3_NAT_dbonly_mixin,
                                                  '_get_router')
        self.mock_get_router = self._mock_get_router.start()
        self.mock_get_router.return_value = {'gw_port_id': self.gw_port_id}
        self._mock_qos_rules = mock.patch.object(self.qos_driver,
                                                 '_qos_rules')
        self.mock_qos_rules = self._mock_qos_rules.start()
        self.fip = {'router_id': 'router_id', 'qos_policy_id': 'qos_policy_id',
                    'floating_network_id': self.network_1,
                    'id': 'fip_id', 'floating_ip_address': '1.2.3.4'}

    def _add_logical_switch(self):
        self.network_1 = 'network_1'
        with self.nb_api.transaction(check_error=True) as txn:
            txn.add(self.nb_api.ls_add(ovn_utils.ovn_name(self.network_1)))

    def _check_rules(self, rules, port_id, network_id, fip_id=None,
                     ip_address=None):
        egress_ovn_rule = self.qos_driver._ovn_qos_rule(
            constants.EGRESS_DIRECTION, rules.get(constants.EGRESS_DIRECTION),
            port_id, network_id, fip_id=fip_id, ip_address=ip_address)
        ingress_ovn_rule = self.qos_driver._ovn_qos_rule(
            constants.INGRESS_DIRECTION,
            rules.get(constants.INGRESS_DIRECTION), port_id, network_id,
            fip_id=fip_id, ip_address=ip_address)

        with self.nb_api.transaction(check_error=True):
            ls = self.qos_driver.nb_idl.lookup(
                'Logical_Switch', ovn_utils.ovn_name(self.network_1))
            self.assertEqual(len(rules), len(ls.qos_rules))
            for rule in ls.qos_rules:
                ref_rule = (egress_ovn_rule if rule.direction == 'from-lport'
                            else ingress_ovn_rule)
                action = {}
                if 'dscp' in ref_rule:
                    action = {'dscp': ref_rule['dscp']}
                bandwidth = {}
                if 'rate' in ref_rule:
                    bandwidth['rate'] = ref_rule['rate']
                    if ref_rule.get('burst'):
                        bandwidth['burst'] = ref_rule['burst']
                self.assertIn(port_id, rule.match)
                self.assertEqual(action, rule.action)
                self.assertEqual(bandwidth, rule.bandwidth)

    def test__update_port_qos_rules(self):
        port = 'port1'

        def update_and_check(qos_rules):
            with self.nb_api.transaction(check_error=True) as txn:
                _qos_rules = copy.deepcopy(qos_rules)
                for direction in constants.VALID_DIRECTIONS:
                    _qos_rules[direction] = _qos_rules.get(direction, {})
                self.mock_qos_rules.return_value = _qos_rules
                self.qos_driver._update_port_qos_rules(
                    txn, port, self.network_1, 'qos1', None)
            self._check_rules(qos_rules, port, self.network_1)

        update_and_check(QOS_RULES_1)
        update_and_check(QOS_RULES_2)
        update_and_check(QOS_RULES_3)
        update_and_check({})

    def _update_fip_and_check(self, fip, qos_rules):
        with self.nb_api.transaction(check_error=True) as txn:
            _qos_rules = copy.deepcopy(qos_rules)
            for direction in constants.VALID_DIRECTIONS:
                _qos_rules[direction] = _qos_rules.get(direction, {})
            self.mock_qos_rules.return_value = _qos_rules
            self.qos_driver.update_floatingip(txn, fip)
        self._check_rules(qos_rules, self.gw_port_id, self.network_1,
                          fip_id='fip_id', ip_address='1.2.3.4')

    def test_create_floatingip(self):
        self._update_fip_and_check(self.fip, QOS_RULES_1)

    def test_update_floatingip(self):
        fip_updated = copy.deepcopy(self.fip)
        fip_updated['qos_policy_id'] = 'another_qos_policy'
        self._update_fip_and_check(self.fip, QOS_RULES_1)
        self._update_fip_and_check(fip_updated, QOS_RULES_2)
        self._update_fip_and_check(fip_updated, QOS_RULES_3)
        self._update_fip_and_check(fip_updated, {})

    def test_delete_floatingip(self):
        self._update_fip_and_check(self.fip, QOS_RULES_1)
        fip_dict = {'floating_network_id': self.fip['floating_network_id'],
                    'id': self.fip['id']}
        self._update_fip_and_check(fip_dict, {})



# Copyright (c) 2015 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import netaddr
from neutron_lib import constants as lib_constants
from neutron_lib.services.qos import constants as qos_consts
from oslo_utils import uuidutils

from neutron.common import ipv6_utils

_uuid = uuidutils.generate_uuid


class FakeDev(object):
    def __init__(self, name):
        self.name = name


def get_ha_interface(ip='169.254.192.1', mac='12:34:56:78:2b:5d'):
    subnet_id = _uuid()
    return {'admin_state_up': True,
            'device_id': _uuid(),
            'device_owner': lib_constants.DEVICE_OWNER_ROUTER_HA_INTF,
            'fixed_ips': [{'ip_address': ip,
                           'prefixlen': 18,
                           'subnet_id': subnet_id}],
            'id': _uuid(),
            'mac_address': mac,
            'name': u'L3 HA Admin port 0',
            'mtu': 1500,
            'network_id': _uuid(),
            'status': u'ACTIVE',
            'subnets': [{'cidr': '169.254.192.0/18',
                         'gateway_ip': '169.254.255.254',
                         'id': subnet_id}],
            'tenant_id': '',
            'agent_id': _uuid(),
            'agent_host': 'aaa',
            'priority': 1}


def prepare_router_data(ip_version=lib_constants.IP_VERSION_4,
                        enable_snat=None, num_internal_ports=1,
                        enable_floating_ip=False, enable_ha=False,
                        extra_routes=False, dual_stack=False, enable_gw=True,
                        v6_ext_gw_with_sub=True,
                        snat_bound_fip=False,
                        enable_pf_floating_ip=False,
                        vrrp_id=None,
                        **kwargs):
    fixed_ips = []
    subnets = []
    gateway_mac = kwargs.get('gateway_mac', 'ca:fe:de:ad:be:ee')
    extra_subnets = []
    for loop_version in (lib_constants.IP_VERSION_4,
                         lib_constants.IP_VERSION_6):
        if (loop_version == lib_constants.IP_VERSION_4 and
                (ip_version == lib_constants.IP_VERSION_4 or dual_stack)):
            ip_address = kwargs.get('ip_address', '19.4.4.4')
            prefixlen = 24
            subnet_cidr = kwargs.get('subnet_cidr', '19.4.4.0/24')
            gateway_ip = kwargs.get('gateway_ip', '19.4.4.1')
            _extra_subnet = {'cidr': '9.4.5.0/24'}
        elif (loop_version == lib_constants.IP_VERSION_6 and
              (ip_version == lib_constants.IP_VERSION_6 or dual_stack) and
              v6_ext_gw_with_sub):
            ip_address = kwargs.get('ip_address', 'fd00::4')
            prefixlen = 64
            subnet_cidr = kwargs.get('subnet_cidr', 'fd00::/64')
            gateway_ip = kwargs.get('gateway_ip', 'fd00::1')
            _extra_subnet = {'cidr': 'fd01::/64'}
        else:
            continue
        subnet_id = _uuid()
        fixed_ips.append({'ip_address': ip_address,
                          'subnet_id': subnet_id,
                          'prefixlen': prefixlen})
        subnets.append({'id': subnet_id,
                        'cidr': subnet_cidr,
                        'gateway_ip': gateway_ip})
        extra_subnets.append(_extra_subnet)
    if not fixed_ips and v6_ext_gw_with_sub:
        raise ValueError("Invalid ip_version: %s" % ip_version)

    router_id = _uuid()
    ex_gw_port = {}
    if enable_gw:
        ex_gw_port = {'id': _uuid(),
                      'mac_address': gateway_mac,
                      'mtu': 1500,
                      'network_id': _uuid(),
                      'fixed_ips': fixed_ips,
                      'subnets': subnets,
                      'extra_subnets': extra_subnets}

    external_gateway_info = {"qos_policy_id": kwargs.get('qos_policy_id')}

    routes = []
    if extra_routes:
        routes = [{'destination': '8.8.8.0/24', 'nexthop': '19.4.4.4'}]

    router = {
        'id': router_id,
        'distributed': False,
        lib_constants.INTERFACE_KEY: [],
        'routes': routes,
        'gw_port': ex_gw_port,
        'external_gateway_info': external_gateway_info}

    router_fips = router.get(lib_constants.FLOATINGIP_KEY, [])
    if enable_floating_ip:
        fip = {'id': _uuid(),
               'port_id': _uuid(),
               'status': 'DOWN',
               'floating_ip_address': '19.4.4.2',
               'fixed_ip_address': '10.0.0.1'}
        qos_policy_id = kwargs.get(qos_consts.QOS_POLICY_ID)
        if qos_policy_id:
            fip[qos_consts.QOS_POLICY_ID] = qos_policy_id
        router_fips.append(fip)

    if snat_bound_fip:
        fip = {'id': _uuid(),
               'port_id': _uuid(),
               'status': 'DOWN',
               'floating_ip_address': '19.4.4.3',
               'fixed_ip_address': '10.0.0.2'}
        qos_policy_id = kwargs.get(qos_consts.QOS_POLICY_ID)
        if qos_policy_id:
            fip[qos_consts.QOS_POLICY_ID] = qos_policy_id
        router_fips.append(fip)
    router[lib_constants.FLOATINGIP_KEY] = router_fips

    pf_fips = []
    if enable_pf_floating_ip:
        fip = {'id': _uuid(),
               'port_id': _uuid(),
               'status': 'DOWN',
               'floating_ip_address': '19.4.4.4',
               'fixed_ip_address': '10.0.0.3'}
        qos_policy_id = kwargs.get(qos_consts.QOS_POLICY_ID)
        if qos_policy_id:
            fip[qos_consts.QOS_POLICY_ID] = qos_policy_id
        pf_fips.append(fip)
    router['_pf_floatingips'] = pf_fips

    router_append_interface(router, count=num_internal_ports,
                            ip_version=ip_version, dual_stack=dual_stack)
    if enable_ha:
        ha_port_ip = kwargs.get('ha_port_ip', '169.254.192.1')
        ha_port_mac = kwargs.get('ha_port_mac', '12:34:56:78:2b:aa')
        router['ha'] = True
        router['ha_vr_id'] = vrrp_id or 1
        router[lib_constants.HA_INTERFACE_KEY] = (
            get_ha_interface(ip=ha_port_ip,
                             mac=ha_port_mac))

    if enable_snat is not None:
        router['enable_snat'] = enable_snat
    return router


def get_subnet_id(port):
    return port['fixed_ips'][0]['subnet_id']


def router_append_interface(router, count=1,
                            ip_version=lib_constants.IP_VERSION_4,
                            ra_mode=None, addr_mode=None, dual_stack=False,
                            same_port=False):
    interfaces = router[lib_constants.INTERFACE_KEY]
    current = sum(
        [netaddr.IPNetwork(subnet['cidr']).version == ip_version
         for p in interfaces for subnet in p['subnets']])

    # If dual_stack=True, create IPv4 and IPv6 subnets on each port
    # If same_port=True, create ip_version number of subnets on a single port
    # Else create just an ip_version subnet on each port
    if dual_stack:
        ip_versions = [lib_constants.IP_VERSION_4, lib_constants.IP_VERSION_6]
    elif same_port:
        ip_versions = [ip_version] * count
        count = 1
    else:
        ip_versions = [ip_version]

    mac_address = netaddr.EUI('ca:fe:de:ad:be:ef')
    mac_address.dialect = netaddr.mac_unix
    for i in range(current, current + count):
        fixed_ips = []
        subnets = []
        for loop_version in ip_versions:
            if (loop_version == lib_constants.IP_VERSION_4 and
                    (ip_version == lib_constants.IP_VERSION_4 or dual_stack)):
                ip_pool = '35.4.%i.4'
                cidr_pool = '35.4.%i.0/24'
                prefixlen = 24
                gw_pool = '35.4.%i.1'
            elif (loop_version == lib_constants.IP_VERSION_6 and
                    (ip_version == lib_constants.IP_VERSION_6 or dual_stack)):
                ip_pool = 'fd01:%x:1::6'
                cidr_pool = 'fd01:%x:1::/64'
                prefixlen = 64
                gw_pool = 'fd01:%x:1::1'
            else:
                continue
            subnet_id = _uuid()
            fixed_ips.append({'ip_address': ip_pool % i,
                              'subnet_id': subnet_id,
                              'prefixlen': prefixlen})
            subnets.append({'id': subnet_id,
                            'cidr': cidr_pool % i,
                            'gateway_ip': gw_pool % i,
                            'ipv6_ra_mode': ra_mode,
                            'ipv6_address_mode': addr_mode})
        if not fixed_ips:
            raise ValueError("Invalid ip_version: %s" % ip_version)

        interfaces.append(
            {'id': _uuid(),
             'mtu': 1500,
             'network_id': _uuid(),
             'admin_state_up': True,
             'fixed_ips': fixed_ips,
             'mac_address': str(mac_address),
             'subnets': subnets})
        mac_address.value += 1


def router_append_subnet(router, count=1,
                         ip_version=lib_constants.IP_VERSION_4,
                         ipv6_subnet_modes=None, interface_id=None,
                         dns_nameservers=None, network_mtu=0):
    if ip_version == lib_constants.IP_VERSION_6:
        subnet_mode_none = {'ra_mode': None, 'address_mode': None}
        if not ipv6_subnet_modes:
            ipv6_subnet_modes = [subnet_mode_none] * count
        elif len(ipv6_subnet_modes) != count:
            ipv6_subnet_modes.extend([subnet_mode_none for i in
                                      range(len(ipv6_subnet_modes), count)])

    if ip_version == lib_constants.IP_VERSION_4:
        ip_pool = '35.4.%i.4'
        cidr_pool = '35.4.%i.0/24'
        prefixlen = 24
        gw_pool = '35.4.%i.1'
    elif ip_version == lib_constants.IP_VERSION_6:
        ip_pool = 'fd01:%x::6'
        cidr_pool = 'fd01:%x::/64'
        prefixlen = 64
        gw_pool = 'fd01:%x::1'
    else:
        raise ValueError("Invalid ip_version: %s" % ip_version)

    interfaces = copy.deepcopy(router.get(lib_constants.INTERFACE_KEY, []))
    if interface_id:
        try:
            interface = next(i for i in interfaces
                             if i['id'] == interface_id)
        except StopIteration:
            raise ValueError("interface_id not found")

        fixed_ips, subnets = interface['fixed_ips'], interface['subnets']
    else:
        interface = None
        fixed_ips, subnets = [], []

    num_existing_subnets = len(subnets)
    for i in range(count):
        subnet_id = _uuid()
        fixed_ips.append(
                {'ip_address': ip_pool % (i + num_existing_subnets),
                 'subnet_id': subnet_id,
                 'prefixlen': prefixlen})
        subnets.append(
                {'id': subnet_id,
                 'cidr': cidr_pool % (i + num_existing_subnets),
                 'gateway_ip': gw_pool % (i + num_existing_subnets),
                 'dns_nameservers': dns_nameservers,
                 'ipv6_ra_mode': ipv6_subnet_modes[i]['ra_mode'],
                 'ipv6_address_mode': ipv6_subnet_modes[i]['address_mode']})

    if interface:
        # Update old interface
        index = interfaces.index(interface)
        interfaces[index].update({'fixed_ips': fixed_ips, 'subnets': subnets})
    else:
        # New interface appended to interfaces list
        mac_address = netaddr.EUI('ca:fe:de:ad:be:ef')
        mac_address.dialect = netaddr.mac_unix
        interfaces.append(
            {'id': _uuid(),
             'mtu': network_mtu,
             'network_id': _uuid(),
             'admin_state_up': True,
             'mac_address': str(mac_address),
             'fixed_ips': fixed_ips,
             'subnets': subnets})

    router[lib_constants.INTERFACE_KEY] = interfaces


def router_append_pd_enabled_subnet(router, count=1, prefix=None):
    if not prefix:
        prefix = lib_constants.PROVISIONAL_IPV6_PD_PREFIX
    interfaces = router[lib_constants.INTERFACE_KEY]
    current = sum(netaddr.IPNetwork(subnet['cidr']).version == 6
                  for p in interfaces for subnet in p['subnets'])

    mac_address = netaddr.EUI('ca:fe:de:ad:be:ef')
    mac_address.dialect = netaddr.mac_unix
    pd_intfs = []
    for i in range(current, current + count):
        subnet_id = _uuid()
        intf = {'id': _uuid(),
                'mtu': 1500,
                'network_id': _uuid(),
                'admin_state_up': True,
                'fixed_ips': [{'ip_address': '::1',
                               'prefixlen': 64,
                               'subnet_id': subnet_id}],
                'mac_address': str(mac_address),
                'subnets': [{'id': subnet_id,
                             'cidr': prefix,
                             'gateway_ip': '::1',
                             'ipv6_ra_mode': lib_constants.IPV6_SLAAC,
                             'subnetpool_id': lib_constants.IPV6_PD_POOL_ID}]}
        interfaces.append(intf)
        pd_intfs.append(intf)
        mac_address.value += 1


def get_unassigned_pd_interfaces(router):
    pd_intfs = []
    for intf in router[lib_constants.INTERFACE_KEY]:
        for subnet in intf['subnets']:
            if (ipv6_utils.is_ipv6_pd_enabled(subnet) and
                    subnet['cidr'] ==
                    lib_constants.PROVISIONAL_IPV6_PD_PREFIX):
                pd_intfs.append(intf)
    return pd_intfs


def get_assigned_pd_interfaces(router):
    pd_intfs = []
    for intf in router[lib_constants.INTERFACE_KEY]:
        for subnet in intf['subnets']:
            if (ipv6_utils.is_ipv6_pd_enabled(subnet) and
                    subnet['cidr'] !=
                    lib_constants.PROVISIONAL_IPV6_PD_PREFIX):
                pd_intfs.append(intf)
    return pd_intfs


def assign_prefix_for_pd_interfaces(router):
    pd_intfs = []
    for ifno, intf in enumerate(router[lib_constants.INTERFACE_KEY]):
        for subnet in intf['subnets']:
            if (ipv6_utils.is_ipv6_pd_enabled(subnet) and
                    subnet['cidr'] ==
                    lib_constants.PROVISIONAL_IPV6_PD_PREFIX):
                subnet['cidr'] = "2001:db8:%d::/64" % ifno
                pd_intfs.append(intf)
    return pd_intfs


def prepare_ext_gw_test(context, ri, dual_stack=False):
    subnet_id = _uuid()
    fixed_ips = [{'subnet_id': subnet_id,
                  'ip_address': '20.0.0.30',
                  'prefixlen': 24}]
    subnets = [{'id': subnet_id,
                'cidr': '20.0.0.0/24',
                'gateway_ip': '20.0.0.1'}]
    if dual_stack:
        subnet_id_v6 = _uuid()
        fixed_ips.append({'subnet_id': subnet_id_v6,
                          'ip_address': '2001:192:168:100::2',
                          'prefixlen': 64})
        subnets.append({'id': subnet_id_v6,
                        'cidr': '2001:192:168:100::/64',
                        'gateway_ip': '2001:192:168:100::1'})
    ex_gw_port = {'fixed_ips': fixed_ips,
                  'subnets': subnets,
                  'extra_subnets': [{'cidr': '172.16.0.0/24'}],
                  'id': _uuid(),
                  'mtu': 1500,
                  'network_id': _uuid(),
                  'mac_address': 'ca:fe:de:ad:be:ef'}
    interface_name = ri.get_external_device_name(ex_gw_port['id'])

    context.device_exists.return_value = True

    return interface_name, ex_gw_port



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import collections
from collections import abc as collections_abc
import copy
import functools
import itertools

from neutron_lib.db import api as db_api
from neutron_lib.db import standard_attr
from neutron_lib import exceptions as n_exc
from neutron_lib.objects import exceptions as o_exc
from neutron_lib.objects.extensions import standardattributes
from oslo_db import exception as obj_exc
from oslo_db.sqlalchemy import enginefacade
from oslo_db.sqlalchemy import utils as db_utils
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import versionutils
from oslo_versionedobjects import base as obj_base
from oslo_versionedobjects import exception as obj_exception
from oslo_versionedobjects import fields as obj_fields
from sqlalchemy import orm
from sqlalchemy.orm import query as sqla_query

from neutron._i18n import _
from neutron.objects.db import api as obj_db_api


LOG = logging.getLogger(__name__)

_NO_DB_MODEL = object()


def get_object_class_by_model(model):
    for obj_class in NeutronObjectRegistry.obj_classes().values():
        obj_class = obj_class[0]
        if getattr(obj_class, 'db_model', _NO_DB_MODEL) is model:
            return obj_class
    raise o_exc.NeutronDbObjectNotFoundByModel(model=model.__name__)


def register_filter_hook_on_model(model, filter_name):
    obj_class = get_object_class_by_model(model)
    obj_class.add_extra_filter_name(filter_name)


class LazyQueryIterator(object):
    def __init__(self, obj_class, lazy_query):
        self.obj_class = obj_class
        self.context = None
        self.query = lazy_query

    def __iter__(self):
        self.results = self.query.all()
        self.i = 0
        return self

    def __next__(self):
        if self.i >= len(self.results):
            raise StopIteration()
        item = self.obj_class._load_object(self.context, self.results[self.i])
        self.i += 1
        return item


class Pager(object):
    '''Pager class

    This class represents a pager object. It is consumed by get_objects to
    specify sorting and pagination criteria.
    '''
    def __init__(self, sorts=None, limit=None, page_reverse=None, marker=None):
        '''Initialize

        :param sorts: A list of (key, direction) tuples.
                      direction: True == ASC, False == DESC
        :param limit: maximum number of items to return
        :param page_reverse: True if sort direction is reversed.
        :param marker: the last item of the previous page; when used, returns
                       next results after the marker resource.
        '''

        self.sorts = sorts
        self.limit = limit
        self.page_reverse = page_reverse
        self.marker = marker

    def to_kwargs(self, context, obj_cls):
        res = {
            attr: getattr(self, attr)
            for attr in ('sorts', 'limit', 'page_reverse')
            if getattr(self, attr) is not None
        }
        if self.marker and self.limit:
            res['marker_obj'] = obj_db_api.get_object(
                obj_cls, context, id=self.marker)
        return res

    def __str__(self):
        return str(self.__dict__)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class NeutronObjectRegistry(obj_base.VersionedObjectRegistry):

    _registry = None

    def __new__(cls, *args, **kwargs):
        # TODO(slaweq): this should be moved back to oslo.versionedobjects
        # lib as soon as bug https://bugs.launchpad.net/neutron/+bug/1731948
        # will be fixed and OVO's registry class will support defining custom
        # registries for objects.

        # NOTE(slaweq): it is overridden method
        # oslo_versionedobjects.base.VersionedObjectRegistry.__new__
        # We need to overwrite it to use separate registry for Neutron's
        # objects.
        # This is necessary to avoid clash in naming objects between Neutron
        # and e.g. os-vif (for example Route or Subnet objects are used in
        # both)
        if not NeutronObjectRegistry._registry:
            NeutronObjectRegistry._registry = object.__new__(
                NeutronObjectRegistry, *args, **kwargs)
            NeutronObjectRegistry._registry._obj_classes = \
                collections.defaultdict(list)
        self = object.__new__(cls, *args, **kwargs)
        self._obj_classes = (
            NeutronObjectRegistry._registry._obj_classes)
        return self


class NeutronObject(obj_base.VersionedObject,
                    obj_base.VersionedObjectDictCompat,
                    obj_base.ComparableVersionedObject,
                    metaclass=abc.ABCMeta):

    synthetic_fields = []
    extra_filter_names = set()

    # To use lazy queries for child objects, you must set the ORM
    # relationship in the db model to 'dynamic'. By default, all
    # children are eager loaded.
    lazy_fields = set()

    def __init__(self, context=None, **kwargs):
        super(NeutronObject, self).__init__(context, **kwargs)
        self._load_synthetic_fields = True
        self.obj_set_defaults()

    def _synthetic_fields_items(self):
        for field in self.synthetic_fields:
            if field in self:
                yield field, getattr(self, field)

    def to_dict(self):
        dict_ = {}
        # not using obj_to_primitive because it skips extra fields
        for name, value in self.items():
            # we have to check if item is in fields because obj_extra_fields
            # is included in self.items()
            if name in self.fields and name not in self.synthetic_fields:
                value = self.fields[name].to_primitive(self, name, value)
            # TODO(ralonsoh): remove once bp/keystone-v3 migration finishes.
            if name == 'tenant_id':
                if ('project_id' in self.fields and
                        not self.obj_attr_is_set('project_id')):
                    continue
            dict_[name] = value
        for field_name, value in self._synthetic_fields_items():
            field = self.fields[field_name]
            if isinstance(field, obj_fields.ListOfObjectsField):
                dict_[field_name] = [obj.to_dict() for obj in value]
            elif isinstance(field, obj_fields.ObjectField):
                dict_[field_name] = (
                    dict_[field_name].to_dict() if value else None)
            else:
                dict_[field_name] = field.to_primitive(self, field_name, value)
        return dict_

    @classmethod
    def is_synthetic(cls, field):
        return field in cls.synthetic_fields

    @classmethod
    def is_object_field(cls, field):
        return (isinstance(cls.fields[field], obj_fields.ListOfObjectsField) or
                isinstance(cls.fields[field], obj_fields.ObjectField))

    @classmethod
    def obj_class_from_name(cls, objname, objver):
        """Returns a class from the registry based on a name and version."""
        # NOTE(slaweq): it is override method
        # oslo_versionedobjects.base.VersionedObject.obj_class_from_name
        # We need to override it to use Neutron's objects registry class
        # (NeutronObjectRegistry) instead of original VersionedObjectRegistry
        # class from oslo_versionedobjects
        # This is necessary to avoid clash in naming objects between Neutron
        # and e.g. os-vif (for example Route or Subnet objects are used in
        # both)
        if objname not in NeutronObjectRegistry.obj_classes():
            LOG.error('Unable to instantiate unregistered object type '
                      '%(objtype)s', dict(objtype=objname))
            raise obj_exception.UnsupportedObjectError(objtype=objname)

        # NOTE(comstud): If there's not an exact match, return the highest
        # compatible version. The objects stored in the class are sorted
        # such that highest version is first, so only set compatible_match
        # once below.
        compatible_match = None

        for objclass in NeutronObjectRegistry.obj_classes()[objname]:
            if objclass.VERSION == objver:
                return objclass
            if (not compatible_match and
                    versionutils.is_compatible(objver, objclass.VERSION)):
                compatible_match = objclass

        if compatible_match:
            return compatible_match

        # As mentioned above, latest version is always first in the list.
        latest_ver = (
            NeutronObjectRegistry.obj_classes()[objname][0].VERSION)
        raise obj_exception.IncompatibleObjectVersion(objname=objname,
                                                      objver=objver,
                                                      supported=latest_ver)

    @classmethod
    def clean_obj_from_primitive(cls, primitive, context=None):
        obj = cls.obj_from_primitive(primitive, context)
        obj.obj_reset_changes()
        return obj

    @classmethod
    def get_object(cls, context, fields=None, **kwargs):
        raise NotImplementedError()

    @classmethod
    def add_extra_filter_name(cls, filter_name):
        """Register filter passed from API layer.

        :param filter_name: Name of the filter passed in the URL

        Filter names are validated in validate_filters() method which by
        default allows filters based on fields' names.  Extensions can create
        new filter names.  Such names must be registered to particular object
        with this method.
        """
        cls.extra_filter_names.add(filter_name)

    @classmethod
    def validate_filters(cls, **kwargs):
        bad_filters = {key for key in kwargs
                       if key not in cls.fields or cls.is_synthetic(key)}
        bad_filters.difference_update(cls.extra_filter_names)
        if bad_filters:
            bad_filters = ', '.join(bad_filters)
            msg = _("'%s' is not supported for filtering") % bad_filters
            raise n_exc.InvalidInput(error_message=msg)

    @classmethod
    @abc.abstractmethod
    def get_objects(cls, context, _pager=None, validate_filters=True,
                    fields=None, **kwargs):
        raise NotImplementedError()

    @classmethod
    def get_values(cls, context, field, validate_filters=True, **kwargs):
        raise NotImplementedError()

    @classmethod
    def _update_objects(cls, objects, values):
        if not isinstance(objects, collections_abc.Sequence):
            objects = (objects, )

        for obj in objects:
            for k, v in values.items():
                setattr(obj, k, v)
            obj.update()
        return len(objects)

    @classmethod
    def update_object(cls, context, values, validate_filters=True, **kwargs):
        obj = cls.get_object(
            context, validate_filters=validate_filters, **kwargs)
        if obj:
            cls._update_objects(obj, values)
            return obj

    @classmethod
    def update_objects(cls, context, values, validate_filters=True, **kwargs):
        objs = cls.get_objects(
            context, validate_filters=validate_filters, **kwargs)
        return cls._update_objects(objs, values)

    @classmethod
    def delete_objects(cls, context, validate_filters=True, **kwargs):
        objs = cls.get_objects(
            context, validate_filters=validate_filters, **kwargs)
        for obj in objs:
            obj.delete()
        return len(objs)

    def create(self):
        raise NotImplementedError()

    def update(self):
        raise NotImplementedError()

    def delete(self):
        raise NotImplementedError()

    @classmethod
    def count(cls, context, validate_filters=True, **kwargs):
        '''Count the number of objects matching filtering criteria.'''
        return len(
            cls.get_objects(
                context, validate_filters=validate_filters, **kwargs))


def _guarantee_rw_subtransaction(func):
    @functools.wraps(func)
    def decorator(self, *args, **kwargs):
        with self.db_context_writer(self.obj_context):
            return func(self, *args, **kwargs)
    return decorator


class DeclarativeObject(abc.ABCMeta):

    def __init__(cls, name, bases, dct):
        super(DeclarativeObject, cls).__init__(name, bases, dct)
        # TODO(ralonsoh): remove once bp/keystone-v3 migration finishes.
        if 'project_id' in cls.fields:
            obj_extra_fields_set = set(cls.obj_extra_fields)
            obj_extra_fields_set.add('tenant_id')
            cls.obj_extra_fields = list(obj_extra_fields_set)
            setattr(cls, 'tenant_id',
                    property(lambda x: x.get('project_id', None)))

        fields_no_update_set = set(cls.fields_no_update)
        for base in itertools.chain([cls], bases):
            keys_set = set()
            if hasattr(base, 'primary_keys'):
                keys_set.update(base.primary_keys)
            if hasattr(base, 'obj_extra_fields'):
                keys_set.update(base.obj_extra_fields)
            for key in keys_set:
                if key in cls.fields or key in cls.obj_extra_fields:
                    fields_no_update_set.add(key)
        cls.fields_no_update = list(fields_no_update_set)

        model = getattr(cls, 'db_model', None)
        if model:
            # generate unique_keys from the model
            if not getattr(cls, 'unique_keys', None):
                cls.unique_keys = []
                obj_field_names = set(cls.fields.keys())
                model_to_obj_translation = {
                    v: k for (k, v) in cls.fields_need_translation.items()}

                keys = db_utils.get_unique_keys(model) or []
                for model_unique_key in keys:
                    obj_unique_key = [model_to_obj_translation.get(key, key)
                                      for key in model_unique_key]
                    if obj_field_names.issuperset(obj_unique_key):
                        cls.unique_keys.append(obj_unique_key)
            cls.create = _guarantee_rw_subtransaction(cls.create)
            cls.update = _guarantee_rw_subtransaction(cls.update)

        if (hasattr(cls, 'has_standard_attributes') and
                cls.has_standard_attributes()):
            setattr(cls, 'standard_attr_id',
                    property(lambda x: x.db_obj.standard_attr_id
                             if x.db_obj else None))
            standardattributes.add_standard_attributes(cls)
            standardattributes.add_tag_filter_names(cls)
        # Instantiate extra filters per class
        cls.extra_filter_names = set(cls.extra_filter_names)
        # TODO(ralonsoh): remove once bp/keystone-v3 migration finishes.
        # add tenant_id filter for objects that have project_id
        if 'project_id' in cls.fields and 'tenant_id' not in cls.fields:
            cls.extra_filter_names.add('tenant_id')

        invalid_fields = [f for f in cls.synthetic_fields
                          if f not in cls.fields]
        if invalid_fields:
            raise o_exc.NeutronObjectValidatorException(fields=invalid_fields)


class NeutronDbObject(NeutronObject, metaclass=DeclarativeObject):

    # should be overridden for all persistent objects
    db_model = None

    # should be overridden for all rbac aware objects
    rbac_db_cls = None

    primary_keys = ['id']

    # 'unique_keys' is a list of unique keys that can be used with get_object
    # instead of 'primary_keys' (e.g. [['key1'], ['key2a', 'key2b']]).
    # By default 'unique_keys' will be inherited from the 'db_model'
    unique_keys = []

    # this is a dict to store the association between the foreign key and the
    # corresponding key in the main table for a synthetic field of a specific
    # class, e.g. port extension has 'port_id' as foreign key, that is
    # associated with the key 'id' of the table Port for the synthetic
    # field of class Port. So foreign_keys = {'Port': {'port_id': 'id'}}.
    # The assumption is the association is the same for all object fields.
    # E.g. all the port extension will use 'port_id' as key.
    foreign_keys = {}

    fields_no_update = []

    # dict with name mapping: {'field_name_in_object': 'field_name_in_db'}
    # It can be used also as DB relationship mapping to synthetic fields name.
    # It is needed to load synthetic fields with one SQL query using side
    # loaded entities.
    # Examples: {'synthetic_field_name': 'relationship_name_in_model'}
    #           {'field_name_in_object': 'field_name_in_db'}
    fields_need_translation = {}

    # obj_extra_fields defines properties that are not part of the model
    # but we want to expose them for easier usage of the object.
    # Handling of obj_extra_fields is in oslo.versionedobjects.
    # The extra fields can be accessed as read only property and are exposed
    # in to_dict()
    # obj_extra_fields = []

    def __init__(self, *args, **kwargs):
        super(NeutronDbObject, self).__init__(*args, **kwargs)
        self._captured_db_model = None

    @property
    def db_obj(self):
        '''Return a database model that persists object data.'''
        return self._captured_db_model

    def _set_lazy_contexts(self, fields, context):
        for field in self.lazy_fields.intersection(fields):
            if isinstance(fields[field], LazyQueryIterator):
                fields[field].context = context

    def from_db_object(self, db_obj):
        fields = self.modify_fields_from_db(db_obj)
        if self.lazy_fields:
            self._set_lazy_contexts(fields, self.obj_context)
        for field in self.fields:
            if field in fields and not self.is_synthetic(field):
                setattr(self, field, fields[field])
        if self._load_synthetic_fields:
            self.load_synthetic_db_fields(db_obj)
        self._captured_db_model = db_obj
        self.obj_reset_changes()

    @classmethod
    def has_standard_attributes(cls):
        return bool(cls.db_model and
                    issubclass(cls.db_model,
                               standard_attr.HasStandardAttributes))

    @classmethod
    def modify_fields_to_db(cls, fields):
        """Modify the fields before data is inserted into DB.

        This method enables to modify the fields and its
        content before data is inserted into DB.

         It uses the fields_need_translation dict with structure:
        {
            'field_name_in_object': 'field_name_in_db'
        }

        :param fields: dict of fields from NeutronDbObject
        :return: modified dict of fields
        """
        for k, v in fields.items():
            if isinstance(v, LazyQueryIterator):
                fields[k] = list(v)
        result = copy.deepcopy(dict(fields))
        for field, field_db in cls.fields_need_translation.items():
            if field in result:
                result[field_db] = result.pop(field)
        return result

    @classmethod
    def _get_lazy_iterator(cls, field, appender_query):
        if field not in cls.lazy_fields:
            raise KeyError(_('Field %s is not a lazy query field') % field)
        n_obj_classes = NeutronObjectRegistry.obj_classes()
        n_obj = n_obj_classes.get(cls.fields[field].objname)
        return LazyQueryIterator(n_obj[0], appender_query)

    @classmethod
    def modify_fields_from_db(cls, db_obj):
        """Modify the fields after data were fetched from DB.

        It uses the fields_need_translation dict with structure:
        {
            'field_name_in_object': 'field_name_in_db'
        }

        :param db_obj: model fetched from database
        :return: modified dict of DB values
        """
        # db models can have declarative proxies that are not exposed into
        # db.keys() so we must fetch data based on object fields definition
        potential_fields = (list(cls.fields.keys()) +
                            list(cls.fields_need_translation.values()))
        # NOTE(ralonsoh): fields dynamically loaded will be represented as
        # ``sqla_query.Query``, because the value is load when needed executing
        # a query to the DB.
        result = {
            field: db_obj[field] for field in potential_fields
            if (db_obj.get(field) is not None and
                not issubclass(db_obj.get(field).__class__, sqla_query.Query))
        }
        for field, field_db in cls.fields_need_translation.items():
            if field_db in result:
                result[field] = result.pop(field_db)
        for k, v in result.items():
            # don't allow sqlalchemy lists to propagate outside
            if isinstance(v, orm.collections.InstrumentedList):
                result[k] = list(v)
            if isinstance(v, orm.dynamic.AppenderQuery):
                result[k] = cls._get_lazy_iterator(k, v)
        return result

    @classmethod
    def _load_object(cls, context, db_obj, fields=None):
        obj = cls(context)

        if fields is not None and len(fields) != 0:
            if len(set(fields).intersection(set(cls.synthetic_fields))) == 0:
                obj._load_synthetic_fields = False

        obj.from_db_object(db_obj)
        return obj

    def obj_load_attr(self, attrname):
        """Set None for nullable fields that has unknown value.

        In case model attribute is not present in database, value stored under
        ``attrname'' field will be unknown. In such cases if the field
        ``attrname'' is a nullable Field return None
        """
        try:
            is_attr_nullable = self.fields[attrname].nullable
        except KeyError:
            return super(NeutronDbObject, self).obj_load_attr(attrname)
        if is_attr_nullable:
            self[attrname] = None

    # TODO(ihrachys) remove once we switch plugin code to enginefacade
    @staticmethod
    def _use_db_facade(context):
        try:
            enginefacade._transaction_ctx_for_context(context)
        except obj_exc.NoEngineContextEstablished:
            return False
        return True

    @classmethod
    def db_context_writer(cls, context):
        """Return read-write session activation decorator."""
        return db_api.CONTEXT_WRITER.using(context)

    @classmethod
    def db_context_reader(cls, context):
        """Return read-only session activation decorator."""
        return db_api.CONTEXT_READER.using(context)

    @classmethod
    def get_object(cls, context, fields=None, **kwargs):
        """Fetch a single object

        Return the first result of given context or None if the result doesn't
        contain any row. Next, convert it to a versioned object.

        :param context:
        :param fields: indicate which fields the caller is interested in
                       using. Note that currently this is limited to
                       avoid loading synthetic fields when possible, and
                       does not affect db queries. Default is None, which
                       is the same as []. Example: ['id', 'name']
        :param kwargs: multiple keys defined by key=value pairs
        :return: single object of NeutronDbObject class or None
        """
        lookup_keys = set(kwargs.keys())
        all_keys = itertools.chain([cls.primary_keys], cls.unique_keys)
        if not any(lookup_keys.issuperset(keys) for keys in all_keys):
            missing_keys = set(cls.primary_keys).difference(lookup_keys)
            raise o_exc.NeutronPrimaryKeyMissing(object_class=cls,
                                                 missing_keys=missing_keys)

        with cls.db_context_reader(context):
            db_obj = obj_db_api.get_object(
                cls, context, **cls.modify_fields_to_db(kwargs))
            if db_obj:
                return cls._load_object(context, db_obj, fields=fields)

    @classmethod
    def get_objects(cls, context, _pager=None, validate_filters=True,
                    fields=None, return_db_obj=False, **kwargs):
        """Fetch a list of objects

        Fetch all results from DB and convert them to versioned objects.

        :param context:
        :param _pager: a Pager object representing advanced sorting/pagination
                       criteria
        :param validate_filters: Raises an error in case of passing an unknown
                                 filter
        :param fields: indicate which fields the caller is interested in
                       using. Note that currently this is limited to
                       avoid loading synthetic fields when possible, and
                       does not affect db queries. Default is None, which
                       is the same as []. Example: ['id', 'name']
        :param return_db_obj: if 'True', the DB object is returned instead of
                              the OVO, saving the conversion time.
        :param kwargs: multiple keys defined by key=value pairs
        :return: list of objects of NeutronDbObject class or empty list
        """
        if validate_filters:
            cls.validate_filters(**kwargs)
        with cls.db_context_reader(context):
            db_objs = obj_db_api.get_objects(
                cls, context, _pager=_pager, **cls.modify_fields_to_db(kwargs))
            if return_db_obj:
                return db_objs

            return [cls._load_object(context, db_obj, fields=fields)
                    for db_obj in db_objs]

    @classmethod
    def get_values(cls, context, field, validate_filters=True, **kwargs):
        """Fetch a list of values of a specific object's field

        Fetch a specific column from DB.

        :param context:
        :param field: a specific field of the object
        :param validate_filters: Raises an error in case of passing an unknown
                                 filter
        :param kwargs: multiple keys defined by key=value pairs
        :return: list of objects of NeutronDbObject class or empty list
        """
        cls._validate_field(field)
        db_field = cls.fields_need_translation.get(field, field)
        if validate_filters:
            cls.validate_filters(**kwargs)
        with cls.db_context_reader(context):
            db_values = obj_db_api.get_values(
                cls, context, db_field, **cls.modify_fields_to_db(kwargs))
            obj = cls(context)
            values = []
            for db_value in db_values:
                value = cls.modify_fields_from_db({
                    db_field: db_value}).get(field)
                value = cls.fields[field].coerce(obj, field, value)
                values.append(value)

            return values

    @classmethod
    def _validate_field(cls, field):
        if field not in cls.fields or cls.is_synthetic(field):
            msg = _("Get value of field '%(field)s' is not supported by "
                    "object '%(object)s'.") % {'field': field, 'object': cls}
            raise n_exc.InvalidInput(error_message=msg)

    @classmethod
    def update_object(cls, context, values, validate_filters=True, **kwargs):
        """Update an object that match filtering criteria from DB.

        :param context:
        :param values: multiple keys to update in matching objects
        :param validate_filters: Raises an error in case of passing an unknown
                                 filter
        :param kwargs: multiple keys defined by key=value pairs
        :return: The updated version of the object
        """
        if validate_filters:
            cls.validate_filters(**kwargs)

        # if we have standard attributes, we will need to fetch records to
        # update revision numbers
        db_obj = None
        if cls.has_standard_attributes():
            return super(NeutronDbObject, cls).update_object(
                context, values, validate_filters=False, **kwargs)
        else:
            with cls.db_context_writer(context):
                db_obj = obj_db_api.update_object(
                    cls, context,
                    cls.modify_fields_to_db(values),
                    **cls.modify_fields_to_db(kwargs))
                return cls._load_object(context, db_obj)

    @classmethod
    def update_objects(cls, context, values, validate_filters=True, **kwargs):
        """Update objects that match filtering criteria from DB.

        :param context:
        :param values: multiple keys to update in matching objects
        :param validate_filters: Raises an error in case of passing an unknown
                                 filter
        :param kwargs: multiple keys defined by key=value pairs
        :return: Number of entries updated
        """
        if validate_filters:
            cls.validate_filters(**kwargs)

        with cls.db_context_writer(context):
            # if we have standard attributes, we will need to fetch records to
            # update revision numbers
            if cls.has_standard_attributes():
                return super(NeutronDbObject, cls).update_objects(
                    context, values, validate_filters=False, **kwargs)
            return obj_db_api.update_objects(
                cls, context,
                cls.modify_fields_to_db(values),
                **cls.modify_fields_to_db(kwargs))

    @classmethod
    def delete_objects(cls, context, validate_filters=True, **kwargs):
        """Delete objects that match filtering criteria from DB.

        :param context:
        :param validate_filters: Raises an error in case of passing an unknown
                                 filter
        :param kwargs: multiple keys defined by key=value pairs
        :return: Number of entries deleted
        """
        if validate_filters:
            cls.validate_filters(**kwargs)
        with cls.db_context_writer(context):
            return obj_db_api.delete_objects(
                cls, context, **cls.modify_fields_to_db(kwargs))

    @classmethod
    def is_accessible(cls, context, db_obj):
        return (context.is_admin or
                context.project_id == db_obj.project_id)

    @staticmethod
    def filter_to_str(value):
        if isinstance(value, list):
            return [str(val) for val in value]
        return str(value)

    @staticmethod
    def filter_to_json_str(value, default=None):
        def _dict_to_json(v):
            return (
                jsonutils.dumps(
                    collections.OrderedDict(
                        sorted(v.items(), key=lambda t: t[0])
                    )
                ) if v else default
            )

        if isinstance(value, list):
            return [_dict_to_json(val) for val in value]
        v = _dict_to_json(value)
        return v

    @staticmethod
    def load_json_from_str(field, default=None):
        value = field or default
        if value:
            value = jsonutils.loads(value)
        return value

    def _get_changed_persistent_fields(self):
        fields = self.obj_get_changes()
        for field in self.synthetic_fields:
            if field in fields:
                del fields[field]
        return fields

    def _validate_changed_fields(self, fields):
        fields = fields.copy()
        forbidden_updates = set(self.fields_no_update) & set(fields.keys())
        if forbidden_updates:
            raise o_exc.NeutronObjectUpdateForbidden(fields=forbidden_updates)

        return fields

    def load_synthetic_db_fields(self, db_obj=None):
        """Load synthetic DB fields

        Load the synthetic fields that are stored in a different table from the
        main object.

        This method doesn't take care of loading synthetic fields that aren't
        stored in the DB, e.g. 'shared' in RBAC policy.
        """
        clsname = self.__class__.__name__

        # TODO(rossella_s) Find a way to handle ObjectFields with
        # subclasses=True
        for field in self.synthetic_fields:
            try:
                field_def = self.fields[field]
                objclasses = NeutronObjectRegistry.obj_classes(
                ).get(field_def.objname)
            except AttributeError:
                # NOTE(rossella_s) this is probably because this field is not
                # an ObjectField
                continue
            if not objclasses:
                # NOTE(rossella_s) some synthetic fields are not handled by
                # this method, for example the ones that have subclasses, see
                # QosRule
                continue
            objclass = objclasses[0]
            foreign_keys = objclass.foreign_keys.get(clsname)
            if not foreign_keys:
                raise o_exc.NeutronSyntheticFieldsForeignKeysNotFound(
                    parent=clsname, child=objclass.__name__)
            if len(foreign_keys.keys()) > 1:
                raise o_exc.NeutronSyntheticFieldMultipleForeignKeys(
                        field=field)

            synthetic_field_db_name = (
                self.fields_need_translation.get(field, field))

            # synth_db_objs can be list, empty list or None, that is why
            # we need 'is not None', because [] is valid case for 'True'
            if isinstance(field_def, obj_fields.ListOfObjectsField):
                synth_db_objs = (db_obj.get(synthetic_field_db_name, None)
                                 if db_obj else None)
                if synth_db_objs is not None:
                    synth_objs = [objclass._load_object(self.obj_context, obj)
                                  for obj in synth_db_objs]
                else:
                    synth_objs = objclass.get_objects(
                        self.obj_context, **{
                            k: getattr(self, v) if v in self else db_obj.get(v)
                            for k, v in foreign_keys.items()})
                setattr(self, field, synth_objs)
            else:
                synth_db_obj = (db_obj.get(synthetic_field_db_name, None)
                                if db_obj else None)
                if synth_db_obj:
                    synth_db_obj = objclass._load_object(self.obj_context,
                                                         synth_db_obj)
                setattr(self, field, synth_db_obj)
            self.obj_reset_changes([field])

    def create(self):
        fields = self._get_changed_persistent_fields()
        with self.db_context_writer(self.obj_context):
            try:
                db_obj = obj_db_api.create_object(
                    self, self.obj_context, self.modify_fields_to_db(fields))
            except obj_exc.DBDuplicateEntry as db_exc:
                raise o_exc.NeutronDbObjectDuplicateEntry(
                    object_class=self.__class__, db_exception=db_exc)

            self.from_db_object(db_obj)

    def _get_composite_keys(self):
        keys = {}
        for key in self.primary_keys:
            keys[key] = getattr(self, key)
        return keys

    def update_fields(self, obj_data, reset_changes=False):
        """Updates fields of an object that are not forbidden to be updated.

        :param obj_data: the full set of object data
        :type obj_data: dict
        :param reset_changes: indicates whether the object's current set of
                              changed fields should be cleared
        :type reset_changes: boolean

        :returns: None
        """
        if reset_changes:
            self.obj_reset_changes()
        for k, v in obj_data.items():
            if k not in self.fields_no_update:
                setattr(self, k, v)

    def update(self):
        updates = self._get_changed_persistent_fields()
        updates = self._validate_changed_fields(updates)

        with self.db_context_writer(self.obj_context):
            db_obj = obj_db_api.update_object(
                self, self.obj_context,
                self.modify_fields_to_db(updates),
                **self.modify_fields_to_db(
                    self._get_composite_keys()))
            self.from_db_object(db_obj)

    def delete(self):
        obj_db_api.delete_object(self, self.obj_context,
                                 **self.modify_fields_to_db(
                                     self._get_composite_keys()))
        self._captured_db_model = None

    @classmethod
    def count(cls, context, validate_filters=True, **kwargs):
        """Count the number of objects matching filtering criteria.

        :param context:
        :param validate_filters: Raises an error in case of passing an unknown
                                 filter
        :param kwargs: multiple keys defined by key=value pairs
        :return: number of matching objects
        """
        if validate_filters:
            cls.validate_filters(**kwargs)
        return obj_db_api.count(
            cls, context, **cls.modify_fields_to_db(kwargs)
        )

    @classmethod
    def objects_exist(cls, context, validate_filters=True, **kwargs):
        """Check if objects are present in DB.

        :param context:
        :param validate_filters: Raises an error in case of passing an unknown
                                 filter
        :param kwargs: multiple keys defined by key=value pairs
        :return: boolean. True if object is present.
        """
        if validate_filters:
            cls.validate_filters(**kwargs)
        # Succeed if at least a single object matches; no need to fetch more
        return bool(obj_db_api.count(
            cls, context, query_limit=1, **cls.modify_fields_to_db(kwargs))
        )



# Copyright 2012 New Dream Network, LLC (DreamHost)
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import collections
import os.path

import eventlet
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import fileutils
import psutil

from neutron.agent.linux import ip_lib
from neutron.agent.linux import utils
from neutron.common import utils as common_utils
from neutron.conf.agent import common as agent_cfg


LOG = logging.getLogger(__name__)


agent_cfg.register_external_process_opts()
agent_cfg.register_process_monitor_opts(cfg.CONF)


class MonitoredProcess(object, metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def active(self):
        """Boolean representing the running state of the process."""

    @abc.abstractmethod
    def enable(self):
        """Enable the service, or respawn the process."""


class ProcessManager(MonitoredProcess):
    """An external process manager for Neutron spawned processes.

    Note: The manager expects uuid to be in cmdline.
    """
    def __init__(self, conf, uuid, namespace=None, service=None,
                 pids_path=None, default_cmd_callback=None,
                 cmd_addl_env=None, pid_file=None, run_as_root=False,
                 custom_reload_callback=None):

        self.conf = conf
        self.uuid = uuid
        self.namespace = namespace
        self.default_cmd_callback = default_cmd_callback
        self.cmd_addl_env = cmd_addl_env
        self.pids_path = pids_path or self.conf.external_pids
        self.pid_file = pid_file
        self.run_as_root = run_as_root or self.namespace is not None
        self.custom_reload_callback = custom_reload_callback
        self.kill_scripts_path = cfg.CONF.AGENT.kill_scripts_path

        if service:
            self.service_pid_fname = 'pid.' + service
            self.service = service
        else:
            self.service_pid_fname = 'pid'
            self.service = 'default-service'

        fileutils.ensure_tree(os.path.dirname(self.get_pid_file_name()),
                              mode=0o755)

    def enable(self, cmd_callback=None, reload_cfg=False, ensure_active=False):
        if not self.active:
            if not cmd_callback:
                cmd_callback = self.default_cmd_callback
            cmd = cmd_callback(self.get_pid_file_name())

            ip_wrapper = ip_lib.IPWrapper(namespace=self.namespace)
            ip_wrapper.netns.execute(cmd, addl_env=self.cmd_addl_env,
                                     run_as_root=self.run_as_root)
        elif reload_cfg:
            self.reload_cfg()
        if ensure_active:
            common_utils.wait_until_true(lambda: self.active)

    def reload_cfg(self):
        if self.custom_reload_callback:
            self.disable(get_stop_command=self.custom_reload_callback)
        else:
            self.disable('HUP')

    def disable(self, sig='9', get_stop_command=None):
        pid = self.pid

        if self.active:
            if get_stop_command:
                cmd = get_stop_command(self.get_pid_file_name())
                ip_wrapper = ip_lib.IPWrapper(namespace=self.namespace)
                ip_wrapper.netns.execute(cmd, addl_env=self.cmd_addl_env,
                                         run_as_root=self.run_as_root,
                                         privsep_exec=True)
            else:
                cmd = self.get_kill_cmd(sig, pid)
                utils.execute(cmd, run_as_root=self.run_as_root,
                              privsep_exec=True)
                # In the case of shutting down, remove the pid file
                if sig == '9':
                    utils.delete_if_exists(self.get_pid_file_name(),
                                           run_as_root=self.run_as_root)
        elif pid:
            LOG.debug('%(service)s process for %(uuid)s pid %(pid)d is stale, '
                      'ignoring signal %(signal)s',
                      {'service': self.service, 'uuid': self.uuid,
                       'pid': pid, 'signal': sig})
        else:
            LOG.debug('No %(service)s process started for %(uuid)s',
                      {'service': self.service, 'uuid': self.uuid})

    def get_kill_cmd(self, sig, pid):
        if self.kill_scripts_path:
            kill_file = "%s-kill" % self.service
            kill_file_path = os.path.join(self.kill_scripts_path, kill_file)
            if os.path.isfile(kill_file_path):
                return [kill_file_path, sig, pid]
        return ['kill', '-%s' % (sig), pid]

    def get_pid_file_name(self):
        """Returns the file name for a given kind of config file."""
        if self.pid_file:
            return self.pid_file
        else:
            return utils.get_conf_file_name(self.pids_path,
                                            self.uuid,
                                            self.service_pid_fname)

    @property
    def pid(self):
        """Last known pid for this external process spawned for this uuid."""
        return utils.get_value_from_file(self.get_pid_file_name(), int)

    @property
    def active(self):
        cmdline = self.cmdline
        return self.uuid in cmdline if cmdline else False

    @property
    def cmdline(self):
        pid = self.pid
        if not pid:
            return
        try:
            return ' '.join(psutil.Process(pid).cmdline())
        except (psutil.NoSuchProcess, psutil.AccessDenied):
            return


ServiceId = collections.namedtuple('ServiceId', ['uuid', 'service'])


class ProcessMonitor(object):

    def __init__(self, config, resource_type):
        """Handle multiple process managers and watch over all of them.

        :param config: oslo config object with the agent configuration.
        :type config: oslo_config.ConfigOpts
        :param resource_type: can be dhcp, router, etc.
        :type resource_type: str
        """
        self._config = config
        self._resource_type = resource_type

        self._monitored_processes = {}

        if self._config.AGENT.check_child_processes_interval:
            self._spawn_checking_thread()

    def register(self, uuid, service_name, monitored_process):
        """Start monitoring a process.

        The given monitored_process will be tied to it's uuid+service_name
        replacing the old one if it existed already.

        The monitored_process should be enabled before registration,
        otherwise ProcessMonitor could try to enable the process itself,
        which could lead to double enable and if unlucky enough, two processes
        running, and also errors in the logs.

        :param uuid: An ID of the resource for which the process is running.
        :param service_name: A logical service name for this process monitor,
                             so the same uuid provided via process manager
                             can reference several different services.
        :param monitored_process: MonitoredProcess we want to monitor.
        """

        service_id = ServiceId(uuid, service_name)
        self._monitored_processes[service_id] = monitored_process

    def unregister(self, uuid, service_name):
        """Stop monitoring a process.

        The uuid+service_name will be removed from the monitored processes.

        The service must be disabled **after** unregistering, otherwise if
        process monitor checks after you disable the process, and before
        you unregister it, the process will be respawned, and left orphaned
        into the system.

        :param uuid: An ID of the resource for which the process is running.
        :param service_name: A logical service name for this process monitor,
                             so the same uuid provided via process manager
                             can reference several different services.
        """

        service_id = ServiceId(uuid, service_name)
        self._monitored_processes.pop(service_id, None)

    def stop(self):
        """Stop the process monitoring.

        This method will stop the monitoring thread, but no monitored
        process will be stopped.
        """
        self._monitor_processes = False

    def _spawn_checking_thread(self):
        self._monitor_processes = True
        eventlet.spawn(self._periodic_checking_thread)

    @lockutils.synchronized("_check_child_processes")
    def _check_child_processes(self):
        # we build the list of keys before iterating in the loop to cover
        # the case where other threads add or remove items from the
        # dictionary which otherwise will cause a RuntimeError
        for service_id in list(self._monitored_processes):
            pm = self._monitored_processes.get(service_id)

            if pm and not pm.active:
                LOG.error("%(service)s for %(resource_type)s "
                          "with uuid %(uuid)s not found. "
                          "The process should not have died",
                          {'service': service_id.service,
                           'resource_type': self._resource_type,
                           'uuid': service_id.uuid})
                self._execute_action(service_id)
            eventlet.sleep(0)

    def _periodic_checking_thread(self):
        while self._monitor_processes:
            eventlet.sleep(self._config.AGENT.check_child_processes_interval)
            eventlet.spawn(self._check_child_processes)

    def _execute_action(self, service_id):
        action = self._config.AGENT.check_child_processes_action
        action_function = getattr(self, "_%s_action" % action)
        action_function(service_id)

    def _respawn_action(self, service_id):
        LOG.warning("Respawning %(service)s for uuid %(uuid)s",
                    {'service': service_id.service,
                     'uuid': service_id.uuid})
        self._monitored_processes[service_id].enable()

    def _exit_action(self, service_id):
        LOG.error("Exiting agent as programmed in check_child_processes_"
                  "actions")
        self._exit_handler(service_id.uuid, service_id.service)

    def _exit_handler(self, uuid, service):
        """This is an exit handler for the ProcessMonitor.

        It will be called if the administrator configured the exit action in
        check_child_processes_actions, and one of our external processes die
        unexpectedly.
        """
        LOG.error("Exiting agent because of a malfunction with the "
                  "%(service)s process identified by uuid %(uuid)s",
                  {'service': service, 'uuid': uuid})
        raise SystemExit(1)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import cProfile
from datetime import datetime
import functools
import io
import pstats

from oslo_config import cfg
from oslo_log import log

from neutron.common import utils
from neutron.conf import profiling as profiling_conf_opts

LOG = log.getLogger(__name__)


CONF = cfg.CONF
profiling_conf_opts.register_profiling_opts()


SORT_BY = 'cumtime'


def profile(f):
    """Decorator to profile Neutron code.

    This will profile the decorated method or function with cProfile and log
    the result.
    """

    @functools.wraps(f)
    def profile_wrapper(*args, **kwargs):

        try:
            if cfg.CONF.enable_code_profiling:
                profid = "%s.%s" % (f.__module__, f.__name__)
                profiler = cProfile.Profile()
                profiler.enable()
                start_time = datetime.now()
            return f(*args, **kwargs)
        finally:
            if cfg.CONF.enable_code_profiling:
                profiler.disable()
                elapsed_time = datetime.now() - start_time
                elapsed_time_ms = (elapsed_time.seconds * 1000.0 +
                                  elapsed_time.microseconds / 1000.0)
                stream = io.StringIO()
                stats = pstats.Stats(profiler, stream=stream).sort_stats(
                    SORT_BY)
                stats.print_stats(cfg.CONF.code_profiling_calls_to_log)
                stats.print_callers(cfg.CONF.code_profiling_calls_to_log)
                stats.print_callees(cfg.CONF.code_profiling_calls_to_log)
                profiler_info = utils.collect_profiler_info()
                if not profiler_info:
                    profiler_info = {'base_id': 'not available',
                                     'parent_id': 'not available'}
                LOG.debug('os-profiler parent trace-id %(parent_id)s '
                          'trace-id %(trace_id)s %(elapsed_time)7d millisecs '
                          'elapsed for %(method)s(*%(args)s, **%(kwargs)s):'
                          '\n%(profiler_data)s',
                          {'parent_id': profiler_info['parent_id'],
                           'trace_id': profiler_info['base_id'],
                           'elapsed_time': elapsed_time_ms,
                           'method': profid,
                           'args': args,
                           'kwargs': kwargs,
                           'profiler_data': stream.getvalue()})
                stream.close()

    return profile_wrapper



# Copyright 2015 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from alembic import op
from oslo_utils import uuidutils
import sqlalchemy as sa

"""network_rbac

Revision ID: 4ffceebfada
Revises: 30018084ec99
Create Date: 2015-06-14 13:12:04.012457

"""

# revision identifiers, used by Alembic.
revision = '4ffceebfada'
down_revision = '30018084ec99'
depends_on = ('8675309a5c4f',)


# A simple model of the networks table with only the fields needed for
# the migration.
network = sa.Table('networks', sa.MetaData(),
                   sa.Column('id', sa.String(length=36), nullable=False),
                   sa.Column('tenant_id', sa.String(length=255)),
                   sa.Column('shared', sa.Boolean(), nullable=False))

networkrbacs = sa.Table(
    'networkrbacs', sa.MetaData(),
    sa.Column('id', sa.String(length=36), nullable=False),
    sa.Column('object_id', sa.String(length=36), nullable=False),
    sa.Column('tenant_id', sa.String(length=255), nullable=True,
              index=True),
    sa.Column('target_tenant', sa.String(length=255), nullable=False),
    sa.Column('action', sa.String(length=255), nullable=False))


def upgrade():
    op.bulk_insert(networkrbacs, get_values())
    op.drop_column('networks', 'shared')
    # the shared column on subnets was just an internal representation of the
    # shared status of the network it was related to. This is now handled by
    # other logic so we just drop it.
    op.drop_column('subnets', 'shared')


def get_values():
    session = sa.orm.Session(bind=op.get_bind())
    values = []
    for row in session.query(network).filter(network.c.shared).all():
        values.append({'id': uuidutils.generate_uuid(), 'object_id': row[0],
                       'tenant_id': row[1], 'target_tenant': '*',
                       'action': 'access_as_shared'})
    # this commit appears to be necessary to allow further operations
    session.commit()
    return values



# Copyright 2015 OpenStack Foundation.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

from oslo_config import cfg
from oslo_log import log as logging

from neutron.agent.metadata import agent
from neutron.common import cache_utils as cache
from neutron.common import config
from neutron.common import utils
from neutron.conf.agent import common as agent_conf
from neutron.conf.agent.metadata import config as meta
from neutron.conf import service as service_conf

LOG = logging.getLogger(__name__)


def main():
    meta.register_meta_conf_opts(meta.SHARED_OPTS)
    meta.register_meta_conf_opts(meta.UNIX_DOMAIN_METADATA_PROXY_OPTS)
    meta.register_meta_conf_opts(meta.METADATA_PROXY_HANDLER_OPTS)
    cache.register_oslo_configs(cfg.CONF)
    agent_conf.register_agent_state_opts_helper(cfg.CONF)
    service_conf.register_service_opts(service_conf.RPC_EXTRA_OPTS, cfg.CONF)

    config.init(sys.argv[1:])
    config.setup_logging()
    utils.log_opt_values(LOG)
    proxy = agent.UnixDomainMetadataProxy(cfg.CONF)
    proxy.run()



# Copyright 2015 IBM Corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.agent.l3.item_allocator import ItemAllocator


class FipPriority(object):
    def __init__(self, index):
        self.index = index

    def __repr__(self):
        return str(self.index)

    def __hash__(self):
        return hash(self.__repr__())

    def __eq__(self, other):
        if isinstance(other, FipPriority):
            return (self.index == other.index)
        else:
            return False

    def __int__(self):
        return int(self.index)


class FipRulePriorityAllocator(ItemAllocator):
    """Manages allocation of floating ips rule priorities.
        IP rule priorities assigned to DVR floating IPs need
        to be preserved over L3 agent restarts.
        This class provides an allocator which saves the priorities
        to a datastore which will survive L3 agent restarts.
    """
    def __init__(self, data_store_path, priority_rule_start,
                 priority_rule_end):
        """Create the necessary pool and create the item allocator
            using ',' as the delimiter and FipRulePriorityAllocator as the
            class type
        """
        pool = set(FipPriority(str(s)) for s in range(priority_rule_start,
                                                      priority_rule_end))

        super(FipRulePriorityAllocator, self).__init__(data_store_path,
                                                       FipPriority,
                                                       pool)



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from oslo_log import versionutils
from oslo_policy import policy

from neutron.conf.policies import base

DEPRECATED_REASON = """
The network API now supports system scope and default roles.
"""

COLLECTION_PATH = '/networks'
RESOURCE_PATH = '/networks/{id}'

ACTION_POST = [
    {'method': 'POST', 'path': COLLECTION_PATH},
]
ACTION_PUT = [
    {'method': 'PUT', 'path': RESOURCE_PATH},
]
ACTION_DELETE = [
    {'method': 'DELETE', 'path': RESOURCE_PATH},
]
ACTION_GET = [
    {'method': 'GET', 'path': COLLECTION_PATH},
    {'method': 'GET', 'path': RESOURCE_PATH},
]


rules = [
    policy.RuleDefault(
        name='external',
        check_str='field:networks:router:external=True',
        description='Definition of an external network'),

    policy.DocumentedRuleDefault(
        name='create_network',
        check_str=base.PROJECT_MEMBER,
        scope_types=['project'],
        description='Create a network',
        operations=ACTION_POST,
        deprecated_rule=policy.DeprecatedRule(
            name='create_network',
            check_str=base.RULE_ANY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='create_network:shared',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description='Create a shared network',
        operations=ACTION_POST,
        deprecated_rule=policy.DeprecatedRule(
            name='create_network:shared',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='create_network:router:external',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description='Create an external network',
        operations=ACTION_POST,
        deprecated_rule=policy.DeprecatedRule(
            name='create_network:router:external',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='create_network:is_default',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description='Specify ``is_default`` attribute when creating a network',
        operations=ACTION_POST,
        deprecated_rule=policy.DeprecatedRule(
            name='create_network:is_default',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='create_network:port_security_enabled',
        check_str=base.PROJECT_MEMBER,
        scope_types=['project'],
        description=(
            'Specify ``port_security_enabled`` '
            'attribute when creating a network'
        ),
        operations=ACTION_POST,
        deprecated_rule=policy.DeprecatedRule(
            name='create_network:port_security_enabled',
            check_str=base.RULE_ANY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='create_network:segments',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description='Specify ``segments`` attribute when creating a network',
        operations=ACTION_POST,
        deprecated_rule=policy.DeprecatedRule(
            name='create_network:segments',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='create_network:provider:network_type',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description=(
            'Specify ``provider:network_type`` '
            'when creating a network'
        ),
        operations=ACTION_POST,
        deprecated_rule=policy.DeprecatedRule(
            name='create_network:provider:network_type',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='create_network:provider:physical_network',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description=(
            'Specify ``provider:physical_network`` '
            'when creating a network'
        ),
        operations=ACTION_POST,
        deprecated_rule=policy.DeprecatedRule(
            name='create_network:provider:physical_network',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='create_network:provider:segmentation_id',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description=(
            'Specify ``provider:segmentation_id`` when creating a network'
        ),
        operations=ACTION_POST,
        deprecated_rule=policy.DeprecatedRule(
            name='create_network:provider:segmentation_id',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),

    policy.DocumentedRuleDefault(
        name='get_network',
        check_str=base.policy_or(
            base.PROJECT_READER,
            'rule:shared',
            'rule:external',
            base.RULE_ADVSVC
        ),
        scope_types=['project'],
        description='Get a network',
        operations=ACTION_GET,
        deprecated_rule=policy.DeprecatedRule(
            name='get_network',
            check_str=base.policy_or(
                base.RULE_ADMIN_OR_OWNER,
                'rule:shared',
                'rule:external',
                base.RULE_ADVSVC),
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='get_network:router:external',
        check_str=base.PROJECT_READER,
        scope_types=['project'],
        description='Get ``router:external`` attribute of a network',
        operations=ACTION_GET,
        deprecated_rule=policy.DeprecatedRule(
            name='get_network:router:external',
            check_str=base.RULE_ANY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='get_network:segments',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description='Get ``segments`` attribute of a network',
        operations=ACTION_GET,
        deprecated_rule=policy.DeprecatedRule(
            name='get_network:segments',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='get_network:provider:network_type',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description='Get ``provider:network_type`` attribute of a network',
        operations=ACTION_GET,
        deprecated_rule=policy.DeprecatedRule(
            name='get_network:provider:network_type',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='get_network:provider:physical_network',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description='Get ``provider:physical_network`` attribute of a network',
        operations=ACTION_GET,
        deprecated_rule=policy.DeprecatedRule(
            name='get_network:provider:physical_network',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='get_network:provider:segmentation_id',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description='Get ``provider:segmentation_id`` attribute of a network',
        operations=ACTION_GET,
        deprecated_rule=policy.DeprecatedRule(
            name='get_network:provider:segmentation_id',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),

    policy.DocumentedRuleDefault(
        name='update_network',
        check_str=base.PROJECT_MEMBER,
        scope_types=['project'],
        description='Update a network',
        operations=ACTION_PUT,
        deprecated_rule=policy.DeprecatedRule(
            name='update_network',
            check_str=base.RULE_ADMIN_OR_OWNER,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='update_network:segments',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description='Update ``segments`` attribute of a network',
        operations=ACTION_PUT,
        deprecated_rule=policy.DeprecatedRule(
            name='update_network:segments',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='update_network:shared',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description='Update ``shared`` attribute of a network',
        operations=ACTION_PUT,
        deprecated_rule=policy.DeprecatedRule(
            name='update_network:shared',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='update_network:provider:network_type',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description='Update ``provider:network_type`` attribute of a network',
        operations=ACTION_PUT,
        deprecated_rule=policy.DeprecatedRule(
            name='update_network:provider:network_type',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='update_network:provider:physical_network',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description=(
            'Update ``provider:physical_network`` '
            'attribute of a network'
        ),
        operations=ACTION_PUT,
        deprecated_rule=policy.DeprecatedRule(
            name='update_network:provider:physical_network',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='update_network:provider:segmentation_id',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description=(
            'Update ``provider:segmentation_id`` '
            'attribute of a network'
        ),
        operations=ACTION_PUT,
        deprecated_rule=policy.DeprecatedRule(
            name='update_network:provider:segmentation_id',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='update_network:router:external',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description='Update ``router:external`` attribute of a network',
        operations=ACTION_PUT,
        deprecated_rule=policy.DeprecatedRule(
            name='update_network:router:external',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='update_network:is_default',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description='Update ``is_default`` attribute of a network',
        operations=ACTION_PUT,
        deprecated_rule=policy.DeprecatedRule(
            name='update_network:is_default',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='update_network:port_security_enabled',
        check_str=base.PROJECT_MEMBER,
        scope_types=['project'],
        description='Update ``port_security_enabled`` attribute of a network',
        operations=ACTION_PUT,
        deprecated_rule=policy.DeprecatedRule(
            name='update_network:port_security_enabled',
            check_str=base.RULE_ADMIN_OR_OWNER,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),

    policy.DocumentedRuleDefault(
        name='delete_network',
        check_str=base.PROJECT_MEMBER,
        scope_types=['project'],
        description='Delete a network',
        operations=ACTION_DELETE,
        deprecated_rule=policy.DeprecatedRule(
            name='delete_network',
            check_str=base.RULE_ADMIN_OR_OWNER,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
]


def list_rules():
    return rules



# Copyright 2020 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures
from ovsdbapp.backend.ovs_idl import connection

from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb import ovsdb_monitor


class OVNIdlConnectionFixture(fixtures.Fixture):
    def __init__(self, idl=None, constr=None, helper=None, timeout=60):
        self.idl = idl or ovsdb_monitor.BaseOvnIdl.from_server(
            constr, helper)
        self.connection = connection.Connection(
            idl=self.idl, timeout=timeout)

    def _setUp(self):
        self.addCleanup(self.stop)
        self.connection.start()

    def stop(self):
        self.connection.stop()



# Copyright 2018 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import portbindings
from neutron_lib import constants
from oslo_utils import uuidutils

from neutron.common import utils as common_utils
from neutron.tests.common.exclusive_resources import ip_network
from neutron.tests.fullstack import base
from neutron.tests.fullstack.resources import environment
from neutron.tests.fullstack.resources import machine
from neutron.tests.unit import testlib_api

load_tests = testlib_api.module_load_tests


class TestPortsRebind(base.BaseFullStackTestCase):

    scenarios = [
        ('Open vSwitch Agent', {
            'l2_agent_type': constants.AGENT_TYPE_OVS,
            'l2_mechdriver_name': 'openvswitch',
        }),
        ('Linux Bridge Agent', {
            'l2_agent_type': constants.AGENT_TYPE_LINUXBRIDGE,
            'l2_mechdriver_name': 'linuxbridge',
        })]

    def setUp(self):
        host_descriptions = [
            environment.HostDescription(
                l2_agent_type=self.l2_agent_type,
                l3_agent=self.use_l3_agent)]
        env = environment.Environment(
            environment.EnvironmentDescription(
                agent_down_time=10),
            host_descriptions)

        super(TestPortsRebind, self).setUp(env)

        self.l2_agent_process = self.environment.hosts[0].l2_agent
        self.l2_agent = self.safe_client.client.list_agents(
            agent_type=self.l2_agent_type)['agents'][0]

        self.tenant_id = uuidutils.generate_uuid()
        self.network = self.safe_client.create_network(self.tenant_id)
        self.subnet = self.safe_client.create_subnet(
            self.tenant_id, self.network['id'], '20.0.0.0/24')

    def _ensure_port_bound(self, port_id):
        port = None

        def port_bound():
            nonlocal port
            port = self.safe_client.client.show_port(port_id)['port']
            return (
                port[portbindings.VIF_TYPE] not in
                    [portbindings.VIF_TYPE_UNBOUND,
                     portbindings.VIF_TYPE_BINDING_FAILED])

        common_utils.wait_until_true(port_bound)
        bound_drivers = {'0': self.l2_mechdriver_name}
        self.assertEqual(bound_drivers,
                         port[portbindings.VIF_DETAILS][
                             portbindings.VIF_DETAILS_BOUND_DRIVERS])

    def _ensure_port_binding_failed(self, port_id):

        def port_binding_failed():
            port = self.safe_client.client.show_port(port_id)['port']
            return (port[portbindings.VIF_TYPE] ==
                    portbindings.VIF_TYPE_BINDING_FAILED)

        common_utils.wait_until_true(port_binding_failed)


class TestVMPortRebind(TestPortsRebind):

    use_l3_agent = False

    def test_vm_port_rebound_when_L2_agent_revived(self):
        """Test scenario

        1. Create port which will be properly bound to host
        2. Stop L2 agent and wait until it will be DEAD
        3. Create another port - it should have "binding_failed"
        4. Turn on L2 agent
        5. Port from p.3 should be bound properly after L2 agent will be UP
        """

        vm_1 = self.useFixture(
            machine.FakeFullstackMachine(
                self.environment.hosts[0],
                self.network['id'],
                self.tenant_id,
                self.safe_client))
        vm_1.block_until_boot()
        self._ensure_port_bound(vm_1.neutron_port['id'])
        vm_1_port = self.safe_client.client.show_port(
            vm_1.neutron_port['id'])['port']

        self.l2_agent_process = self.environment.hosts[0].l2_agent
        self.l2_agent = self.safe_client.client.list_agents(
            agent_type=self.l2_agent_type)['agents'][0]

        self.l2_agent_process.stop()
        self._wait_until_agent_down(self.l2_agent['id'])

        vm_2 = self.useFixture(
            machine.FakeFullstackMachine(
                self.environment.hosts[0],
                self.network['id'],
                self.tenant_id,
                self.safe_client))
        self._ensure_port_binding_failed(vm_2.neutron_port['id'])
        vm_2_port = self.safe_client.client.show_port(
            vm_2.neutron_port['id'])['port']

        # check if vm_1 port is still bound as it was before
        self._ensure_port_bound(vm_1.neutron_port['id'])
        # and that revision number of vm_1's port wasn't changed
        self.assertEqual(
            vm_1_port['revision_number'],
            self.safe_client.client.show_port(
                vm_1_port['id'])['port']['revision_number'])

        self.l2_agent_process.start()
        self._wait_until_agent_up(self.l2_agent['id'])

        self._ensure_port_bound(vm_2_port['id'])


class TestRouterPortRebind(TestPortsRebind):

    use_l3_agent = True

    def setUp(self):
        super(TestRouterPortRebind, self).setUp()

        self.tenant_id = uuidutils.generate_uuid()
        self.ext_net = self.safe_client.create_network(
            self.tenant_id, external=True)
        ext_cidr = self.useFixture(
            ip_network.ExclusiveIPNetwork(
                "240.0.0.0", "240.255.255.255", "24")).network
        self.safe_client.create_subnet(
            self.tenant_id, self.ext_net['id'], ext_cidr)
        self.router = self.safe_client.create_router(
            self.tenant_id, external_network=self.ext_net['id'])

    def test_vm_port_rebound_when_L2_agent_revived(self):
        """Test scenario

        1. Ensure that router gateway port is bound properly
        2. Stop L2 agent and wait until it will be DEAD
        3. Create router interface and check that it's port is "binding_failed"
        4. Turn on L2 agent
        5. Router's port created in p.3 should be now bound properly
        """

        if self.l2_agent_type == constants.AGENT_TYPE_LINUXBRIDGE:
            self.skipTest("Bug 1798085")

        gw_port = self.safe_client.client.list_ports(
            device_id=self.router['id'],
            device_owner=constants.DEVICE_OWNER_ROUTER_GW)['ports'][0]
        self._ensure_port_bound(gw_port['id'])

        self.l2_agent_process.stop()
        self._wait_until_agent_down(self.l2_agent['id'])

        router_interface_info = self.safe_client.add_router_interface(
            self.router['id'], self.subnet['id'])
        self._ensure_port_binding_failed(router_interface_info['port_id'])

        self.l2_agent_process.start()
        self._wait_until_agent_up(self.l2_agent['id'])

        self._ensure_port_bound(router_interface_info['port_id'])



# Copyright 2015 OpenStack Foundation.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools
import shlex

from neutron.conf.agent.metadata import config as meta_conf
from oslo_config import cfg
from oslo_privsep import priv_context

from neutron._i18n import _

OVS_OPTS = [
    cfg.StrOpt('ovsdb_connection',
               default='unix:/usr/local/var/run/openvswitch/db.sock',
               help=_('The connection string for the native OVSDB backend.\n'
                      'Use tcp:IP:PORT for TCP connection.\n'
                      'Use unix:FILE for unix domain socket connection.')),
    cfg.IntOpt('ovsdb_connection_timeout',
               default=180,
               help=_('Timeout in seconds for the OVSDB '
                      'connection transaction'))
]


def register_meta_conf_opts(opts, cfg=cfg.CONF, group=None):
    cfg.register_opts(opts, group=group)


def list_metadata_agent_opts():
    return [
        ('DEFAULT',
         itertools.chain(
             meta_conf.SHARED_OPTS,
             meta_conf.METADATA_PROXY_HANDLER_OPTS,
             meta_conf.UNIX_DOMAIN_METADATA_PROXY_OPTS)
         ),
        ('ovs', OVS_OPTS)
    ]


def get_root_helper(conf):
    return conf.AGENT.root_helper


def setup_privsep():
    priv_context.init(root_helper=shlex.split(get_root_helper(cfg.CONF)))



# Copyright (C) 2018 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections import namedtuple
from unittest import mock

from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib import context
from neutron_lib import exceptions as lib_exc
from neutron_lib.exceptions import l3 as lib_l3_exc
from neutron_lib.objects import exceptions as obj_exc
from neutron_lib.plugins import constants as lib_plugin_conts
from neutron_lib.plugins import directory
from oslo_config import cfg

from neutron.api.rpc.callbacks.consumer import registry as cons_registry
from neutron.api.rpc.callbacks import events as rpc_events
from neutron.api.rpc.callbacks.producer import registry as prod_registry
from neutron.api.rpc.callbacks import resource_manager
from neutron.api.rpc.handlers import resources_rpc
from neutron.db import db_base_plugin_v2
from neutron.db import l3_db
from neutron import manager
from neutron.objects import base as obj_base
from neutron.objects import port_forwarding
from neutron.objects import router
from neutron.services.portforwarding.common import exceptions as pf_exc
from neutron.services.portforwarding import constants as pf_consts
from neutron.services.portforwarding import pf_plugin
from neutron.tests.unit import testlib_api


DB_PLUGIN_KLASS = 'neutron.db.db_base_plugin_v2.NeutronDbPluginV2'


class TestPortForwardingPlugin(testlib_api.SqlTestCase):

    def setUp(self):
        super(TestPortForwardingPlugin, self).setUp()

        with mock.patch.object(
                resource_manager.ResourceCallbacksManager, '_singleton',
                new_callable=mock.PropertyMock(return_value=False)):

            self.cons_mgr = resource_manager.ConsumerResourceCallbacksManager()
            self.prod_mgr = resource_manager.ProducerResourceCallbacksManager()
            for mgr in (self.cons_mgr, self.prod_mgr):
                mgr.clear()

        mock.patch.object(
            cons_registry, '_get_manager', return_value=self.cons_mgr).start()

        mock.patch.object(
            prod_registry, '_get_manager', return_value=self.prod_mgr).start()
        self.setup_coreplugin(load_plugins=False)

        mock.patch('neutron_lib.callbacks.registry.publish').start()
        mock.patch('neutron.objects.db.api.create_object').start()
        mock.patch('neutron.objects.db.api.update_object').start()
        mock.patch('neutron.objects.db.api.delete_object').start()
        mock.patch('neutron.objects.db.api.get_object').start()
        # We don't use real models as per mocks above. We also need to mock-out
        # methods that work with real data types
        mock.patch(
            'neutron.objects.base.NeutronDbObject.modify_fields_from_db'
        ).start()
        mock.patch.object(router.FloatingIP, 'from_db_object').start()

        cfg.CONF.set_override("core_plugin", DB_PLUGIN_KLASS)
        cfg.CONF.set_override("service_plugins", ["router", "port_forwarding"])

        manager.init()
        self.pf_plugin = directory.get_plugin(lib_plugin_conts.PORTFORWARDING)
        self.ctxt = context.Context('admin', 'fake_tenant')
        mock.patch.object(self.ctxt.session, 'refresh').start()
        mock.patch.object(self.ctxt.session, 'expunge').start()

    @mock.patch.object(port_forwarding.PortForwarding, 'get_object')
    def test_get_floatingip_port_forwarding(self, get_object_mock):
        self.pf_plugin.get_floatingip_port_forwarding(
            self.ctxt, 'pf_id', 'test-fip-id', fields=None)
        get_object_mock.assert_called_once_with(self.ctxt, id='pf_id')

    @mock.patch.object(port_forwarding.PortForwarding, 'get_object',
                       return_value=None)
    def test_negative_get_floatingip_port_forwarding(self, get_object_mock):
        self.assertRaises(
            pf_exc.PortForwardingNotFound,
            self.pf_plugin.get_floatingip_port_forwarding,
            self.ctxt, 'pf_id', 'test-fip-id', fields=None)

    @mock.patch.object(port_forwarding.PortForwarding, 'get_objects')
    def test_get_floatingip_port_forwardings(self, get_objects_mock):
        self.pf_plugin.get_floatingip_port_forwardings(self.ctxt)
        get_objects_mock.assert_called_once_with(
            self.ctxt, _pager=mock.ANY, floatingip_id=None)

    @mock.patch.object(registry, 'publish')
    @mock.patch.object(resources_rpc.ResourcesPushRpcApi, 'push')
    @mock.patch.object(port_forwarding.PortForwarding, 'get_object')
    @mock.patch.object(port_forwarding.PortForwarding, 'get_objects')
    @mock.patch.object(router.FloatingIP, 'get_object')
    def test_delete_floatingip_port_forwarding(
            self, fip_get_object_mock, pf_get_objects_mock,
            pf_get_object_mock, push_api_mock, mock_registry_publish):

        # After delete, not empty resource list
        pf_get_objects_mock.return_value = [mock.Mock(id='pf_id'),
                                            mock.Mock(id='pf_id2')]
        pf_obj = mock.Mock(id='pf_id', floatingip_id='fip_id')
        pf_get_object_mock.return_value = pf_obj
        self.pf_plugin.delete_floatingip_port_forwarding(
            self.ctxt, 'pf_id', 'fip_id')
        pf_get_objects_mock.assert_called_once_with(
            self.ctxt, floatingip_id='fip_id')
        pf_obj.delete.assert_called()
        push_api_mock.assert_called_once_with(
            self.ctxt, mock.ANY, rpc_events.DELETED)
        mock_registry_publish.assert_called_once_with(
            pf_consts.PORT_FORWARDING,
            events.AFTER_DELETE, self.pf_plugin, payload=mock.ANY)

        # After delete, empty resource list
        pf_get_objects_mock.reset_mock()
        pf_get_object_mock.reset_mock()
        push_api_mock.reset_mock()
        mock_registry_publish.reset_mock()
        pf_obj = mock.Mock(id='need_to_delete_pf_id', floatingip_id='fip_id')
        fip_obj = mock.Mock(id='fip_id')
        fip_get_object_mock.return_value = fip_obj
        pf_get_object_mock.return_value = pf_obj
        pf_get_objects_mock.return_value = [
            mock.Mock(id='need_to_delete_pf_id')]

        self.pf_plugin.delete_floatingip_port_forwarding(
            self.ctxt, 'need_to_delete_pf_id', 'fip_id')
        pf_get_objects_mock.assert_called_once_with(
            self.ctxt, floatingip_id='fip_id')
        pf_get_object_mock.assert_called_once_with(
            self.ctxt, id='need_to_delete_pf_id')
        fip_obj.update_fields.assert_called_once_with({'router_id': None})
        fip_obj.update.assert_called()
        push_api_mock.assert_called_once_with(
            self.ctxt, mock.ANY, rpc_events.DELETED)
        mock_registry_publish.assert_called_once_with(
            pf_consts.PORT_FORWARDING,
            events.AFTER_DELETE, self.pf_plugin, payload=mock.ANY)

    @mock.patch.object(port_forwarding.PortForwarding, 'get_object')
    def test_negative_delete_floatingip_port_forwarding(
            self, pf_get_object_mock):
        pf_get_object_mock.return_value = None
        self.assertRaises(
            pf_exc.PortForwardingNotFound,
            self.pf_plugin.delete_floatingip_port_forwarding,
            self.ctxt, 'pf_id', floatingip_id='fip_id')

    @mock.patch.object(port_forwarding.PortForwarding, 'get_objects')
    @mock.patch.object(db_base_plugin_v2.NeutronDbPluginV2, 'get_port')
    @mock.patch.object(registry, 'publish')
    @mock.patch.object(resources_rpc.ResourcesPushRpcApi, 'push')
    @mock.patch.object(port_forwarding.PortForwarding, 'get_object')
    def test_update_floatingip_port_forwarding(
            self, mock_pf_get_object, mock_rpc_push, mock_registry_publish,
            mock_get_port, mock_pf_get_objects):
        pf_input = {
            'port_forwarding':
                {'port_forwarding': {
                    'internal_ip_address': '1.1.1.1',
                    'floatingip_id': 'fip_id'}},
            'floatingip_id': 'fip_id'}
        pf_obj = mock.Mock()
        pf_obj.internal_ip_address = "10.0.0.1"
        mock_pf_get_object.return_value = pf_obj
        port_dict = {'id': 'ID', 'fixed_ips': [{"subnet_id": "test-subnet-id",
                                                "ip_address": "10.0.0.1"}]}
        mock_get_port.return_value = port_dict
        mock_pf_get_objects.return_value = []
        self.pf_plugin.update_floatingip_port_forwarding(
            self.ctxt, 'pf_id', **pf_input)
        mock_pf_get_object.assert_called_once_with(self.ctxt, id='pf_id')
        self.assertTrue(pf_obj.update_fields)
        self.assertTrue(pf_obj.update)
        mock_rpc_push.assert_called_once_with(
            self.ctxt, mock.ANY, rpc_events.UPDATED)
        mock_registry_publish.assert_called_once_with(
            pf_consts.PORT_FORWARDING,
            events.AFTER_UPDATE, self.pf_plugin, payload=mock.ANY)

    @mock.patch.object(port_forwarding.PortForwarding, 'get_objects')
    @mock.patch.object(db_base_plugin_v2.NeutronDbPluginV2, 'get_port')
    def test_check_port_forwarding_update(self, mock_get_port,
                                          mock_pf_get_objects):
        port_dict = {'fixed_ips': [{"ip_address": "10.0.0.1"}]}
        mock_get_port.return_value = port_dict
        other_pf_obj = mock.Mock(id='cmp_obj_id')
        pf_obj = mock.Mock(id='pf_obj_id', internal_port_id='int_port_id',
                           internal_ip_address='10.0.0.1')
        mock_pf_get_objects.return_value = [pf_obj, other_pf_obj]
        self.pf_plugin._check_port_forwarding_update(self.ctxt, pf_obj)
        mock_get_port.assert_called_once_with(self.ctxt, 'int_port_id')
        mock_pf_get_objects.assert_called_once_with(self.ctxt,
            floatingip_id=pf_obj.floatingip_id, protocol=pf_obj.protocol)

    @mock.patch.object(port_forwarding.PortForwarding, 'get_objects')
    @mock.patch.object(db_base_plugin_v2.NeutronDbPluginV2, 'get_port')
    def test_check_port_forwarding_update_invalid_address(
            self, mock_get_port, mock_pf_get_objects):
        port_dict = {'fixed_ips': [{"ip_address": "10.0.0.1"}]}
        mock_get_port.return_value = port_dict
        pf_obj = mock.Mock(id='pf_obj_id', internal_port_id='int_port_id',
                           internal_ip_address="99.99.99.99")
        self.assertRaisesRegex(
            lib_exc.BadRequest,
            "not correspond to an address on the internal port",
            self.pf_plugin._check_port_forwarding_update,
            self.ctxt, pf_obj)
        mock_get_port.assert_called_once_with(self.ctxt, 'int_port_id')
        mock_pf_get_objects.assert_not_called()

    @mock.patch.object(port_forwarding.PortForwarding, 'get_objects')
    @mock.patch.object(db_base_plugin_v2.NeutronDbPluginV2, 'get_port')
    def test_check_port_forwarding_update_invalid_external(
            self, mock_get_port, mock_pf_get_objects):
        port_dict = {'fixed_ips': [{"ip_address": "10.0.0.1"}]}
        mock_get_port.return_value = port_dict
        pf_obj_dict = {'id': 'pf_obj_one',
                       'floating_ip_address': 'same_fip_addr',
                       'external_port': 'same_ext_port'}
        other_pf_obj = mock.Mock(**pf_obj_dict)
        pf_obj_dict.update(id='pf_obj_two', internal_ip_address='10.0.0.1')
        pf_obj = mock.Mock(**pf_obj_dict)
        mock_pf_get_objects.return_value = [pf_obj, other_pf_obj]
        self.assertRaisesRegex(
            lib_exc.BadRequest,
            "already exist.*same_fip_addr",
            self.pf_plugin._check_port_forwarding_update,
            self.ctxt, pf_obj)
        mock_get_port.assert_called_once_with(self.ctxt, mock.ANY)
        mock_pf_get_objects.assert_called_once()

    @mock.patch.object(port_forwarding.PortForwarding, 'get_objects')
    @mock.patch.object(db_base_plugin_v2.NeutronDbPluginV2, 'get_port')
    def test_check_port_forwarding_update_invalid_internal(
            self, mock_get_port, mock_pf_get_objects):
        same_internal_ip = "10.0.0.10"
        port_dict = {'fixed_ips': [{"ip_address": same_internal_ip}]}
        mock_get_port.return_value = port_dict
        pf_obj_dict = {'id': 'pf_obj_one',
                       'internal_port_id': 'same_int_port_id',
                       'internal_ip_address': same_internal_ip,
                       'internal_port': 'same_int_port'}
        other_pf_obj = mock.Mock(**pf_obj_dict)
        pf_obj_dict.update(id='pf_obj_two')
        pf_obj = mock.Mock(**pf_obj_dict)
        mock_pf_get_objects.return_value = [pf_obj, other_pf_obj]
        self.assertRaisesRegex(
            lib_exc.BadRequest,
            "already exist.*{}".format(same_internal_ip),
            self.pf_plugin._check_port_forwarding_update,
            self.ctxt, pf_obj)
        mock_get_port.assert_called_once_with(self.ctxt, 'same_int_port_id')
        mock_pf_get_objects.assert_called_once()

    @mock.patch.object(port_forwarding.PortForwarding, 'get_object')
    def test_negative_update_floatingip_port_forwarding(
            self, mock_pf_get_object):
        pf_input = {
            'port_forwarding':
                {'port_forwarding': {
                    'internal_ip_address': '1.1.1.1',
                    'floatingip_id': 'fip_id'}},
            'floatingip_id': 'fip_id'}
        mock_pf_get_object.return_value = None
        self.assertRaises(
            pf_exc.PortForwardingNotFound,
            self.pf_plugin.update_floatingip_port_forwarding,
            self.ctxt, 'pf_id', **pf_input)

    @mock.patch.object(pf_plugin.PortForwardingPlugin,
                       '_check_port_has_binding_floating_ip')
    @mock.patch.object(obj_base.NeutronDbObject, 'update_objects')
    @mock.patch.object(registry, 'publish')
    @mock.patch.object(resources_rpc.ResourcesPushRpcApi, 'push')
    @mock.patch.object(pf_plugin.PortForwardingPlugin, '_check_router_match')
    @mock.patch.object(pf_plugin.PortForwardingPlugin,
                       '_find_a_router_for_fip_port_forwarding',
                       return_value='target_router_id')
    @mock.patch.object(router.FloatingIP, 'get_object')
    @mock.patch('neutron.objects.port_forwarding.PortForwarding')
    def test_create_floatingip_port_forwarding(
            self, mock_port_forwarding, mock_fip_get_object, mock_find_router,
            mock_check_router_match, mock_push_api, mock_registry_publish,
            mock_update_objects, mock_check_bind_fip):
        # Update fip
        pf_input = {
            'port_forwarding':
                {'port_forwarding': {
                    'internal_ip_address': '1.1.1.1',
                    'floatingip_id': 'fip_id'}},
            'floatingip_id': 'fip_id'}
        pf_obj = mock.Mock()
        fip_obj = mock.Mock()
        mock_port_forwarding.return_value = pf_obj
        mock_fip_get_object.return_value = fip_obj
        fip_obj.router_id = ''
        fip_obj.fixed_port_id = ''
        self.pf_plugin.create_floatingip_port_forwarding(
            self.ctxt, **pf_input)
        mock_port_forwarding.assert_called_once_with(
            self.ctxt, **pf_input['port_forwarding']['port_forwarding'])
        self.assertTrue(mock_update_objects.called)
        self.assertTrue(pf_obj.create.called)
        mock_push_api.assert_called_once_with(
            self.ctxt, mock.ANY, rpc_events.CREATED)
        mock_registry_publish.assert_called_once_with(
            pf_consts.PORT_FORWARDING,
            events.AFTER_CREATE, self.pf_plugin, payload=mock.ANY)

        # Not update fip
        pf_obj.reset_mock()
        fip_obj.reset_mock()
        mock_port_forwarding.reset_mock()
        mock_update_objects.reset_mock()
        mock_push_api.reset_mock()
        mock_registry_publish.reset_mock()
        mock_port_forwarding.return_value = pf_obj
        fip_obj.router_id = 'router_id'
        fip_obj.fixed_port_id = ''
        self.pf_plugin.create_floatingip_port_forwarding(
            self.ctxt, **pf_input)
        mock_port_forwarding.assert_called_once_with(
            self.ctxt, **pf_input['port_forwarding']['port_forwarding'])
        self.assertTrue(pf_obj.create.called)
        self.assertFalse(mock_update_objects.called)
        mock_push_api.assert_called_once_with(
            self.ctxt, mock.ANY, rpc_events.CREATED)
        mock_registry_publish.assert_called_once_with(
            pf_consts.PORT_FORWARDING,
            events.AFTER_CREATE, self.pf_plugin, payload=mock.ANY)

    @mock.patch.object(pf_plugin.PortForwardingPlugin,
                       '_check_port_has_binding_floating_ip')
    @mock.patch.object(pf_plugin.PortForwardingPlugin,
                       '_find_existing_port_forwarding')
    @mock.patch.object(pf_plugin.PortForwardingPlugin,
                       '_check_router_match')
    @mock.patch.object(pf_plugin.PortForwardingPlugin,
                       '_find_a_router_for_fip_port_forwarding',
                       return_value='target_router_id')
    @mock.patch.object(router.FloatingIP, 'get_object')
    @mock.patch('neutron.objects.port_forwarding.PortForwarding')
    def test_negative_create_floatingip_port_forwarding(
            self, mock_port_forwarding, mock_fip_get_object,
            mock_find_router,
            mock_check_router_match, mock_try_find_exist,
            mock_check_bind_fip):
        pf_input = {
            'port_forwarding': {
                'internal_ip_address': '1.1.1.1',
                'floatingip_id': 'fip_id'}}
        pf_obj = mock.Mock()
        fip_obj = mock.Mock()
        mock_port_forwarding.return_value = pf_obj
        mock_fip_get_object.return_value = fip_obj
        fip_obj.fixed_port_id = ''

        pf_obj.create.side_effect = obj_exc.NeutronDbObjectDuplicateEntry(
            mock.Mock(), mock.Mock())
        mock_try_find_exist.return_value = ('pf_obj', 'conflict_param')
        self.assertRaises(
            lib_exc.BadRequest,
            self.pf_plugin.create_floatingip_port_forwarding,
            self.ctxt, 'fip_id', pf_input)

    @mock.patch.object(pf_plugin.PortForwardingPlugin,
                       '_get_internal_ip_subnet')
    @mock.patch.object(l3_db.L3_NAT_dbonly_mixin, 'get_router_for_floatingip')
    @mock.patch.object(db_base_plugin_v2.NeutronDbPluginV2, 'get_port')
    @mock.patch.object(db_base_plugin_v2.NeutronDbPluginV2, 'get_subnet')
    def test_negative_find_a_router_for_fip_port_forwarding(
            self, mock_get_subnet, mock_get_port, mock_get_router,
            mock_get_ip_subnet):
        fip_obj = mock.Mock()
        pf_dict = {'internal_port_id': 'internal_neutron_port',
                   'internal_ip_address': '10.0.0.1'}
        port_dict = {'id': 'ID', 'fixed_ips': [{"subnet_id": "test-subnet-id",
                                                "ip_address": "10.0.0.1"}]}
        mock_get_port.return_value = port_dict
        mock_get_ip_subnet.return_value = None
        self.assertRaises(
            lib_exc.BadRequest,
            self.pf_plugin._find_a_router_for_fip_port_forwarding,
            self.ctxt, pf_dict, fip_obj)
        self.assertTrue(not mock_get_subnet.called)

        mock_get_ip_subnet.return_value = 'internal_subnet_id'

        mock_get_router.side_effect = (
            lib_l3_exc.ExternalGatewayForFloatingIPNotFound(
                external_network_id=mock.Mock(),
                subnet_id=mock.Mock(), port_id=mock.Mock()))
        self.assertRaises(
            lib_exc.BadRequest,
            self.pf_plugin._find_a_router_for_fip_port_forwarding,
            self.ctxt, pf_dict, fip_obj)
        self.assertTrue(mock_get_subnet.called)

        ipv6_port_dict = {'id': 'ID',
                          'fixed_ips': [{"subnet_id": "test-subnet-id",
                                         "ip_address": "1::1"}]}
        mock_get_port.return_value = ipv6_port_dict
        self.assertRaises(
            lib_exc.BadRequest,
            self.pf_plugin._find_a_router_for_fip_port_forwarding,
            self.ctxt, pf_dict, fip_obj)

    @mock.patch.object(port_forwarding.PortForwarding, 'get_objects')
    def test_negative_check_router_match(self, mock_pf_get_objects):
        pf_dict = {
            'internal_port_id': 'internal_neutron_port',
            'internal_ip_address': 'internal_fixed_ip',
            'internal_port': 'internal protocol port num'}
        fip_obj = mock.Mock()
        mock_pf_get_objects.return_value = ['Exist port forwardings']
        router_id = 'selected router id'
        self.assertRaises(lib_exc.BadRequest,
                          self.pf_plugin._check_router_match,
                          self.ctxt, fip_obj, router_id, pf_dict)

        mock_pf_get_objects.return_value = []
        self.assertRaises(lib_exc.BadRequest,
                          self.pf_plugin._check_router_match,
                          self.ctxt, fip_obj, router_id, pf_dict)

    @mock.patch.object(router.FloatingIP, 'get_objects')
    def test_create_floatingip_port_forwarding_port_in_use(
            self, mock_fip_get_objects):
        pf_input = {
            'port_forwarding':
                {'port_forwarding': {
                    'internal_ip_address': '1.1.1.1',
                    'internal_port_id': 'internal_neutron_port',
                    'floatingip_id': 'fip_id_1'}},
            'floatingip_id': 'fip_id_1'}
        fip_obj = mock.Mock(floating_ip_address="10.10.10.10")
        mock_fip_get_objects.return_value = [fip_obj]
        self.assertRaises(pf_exc.PortHasBindingFloatingIP,
                          self.pf_plugin.create_floatingip_port_forwarding,
                          self.ctxt, **pf_input)

    @mock.patch.object(router.FloatingIP, 'get_objects')
    def test_update_floatingip_port_forwarding_port_in_use(
            self, mock_fip_get_objects):
        pf_input = {
            'port_forwarding':
                {'port_forwarding': {
                    'internal_ip_address': '1.1.1.1',
                    'internal_port_id': 'internal_neutron_port',
                    'floatingip_id': 'fip_id_2'}}}
        fip_obj = mock.Mock(floating_ip_address="10.10.10.11")
        mock_fip_get_objects.return_value = [fip_obj]
        self.assertRaises(pf_exc.PortHasBindingFloatingIP,
                          self.pf_plugin.update_floatingip_port_forwarding,
                          self.ctxt, 'fake-pf-id', 'fip_id_2', **pf_input)

    def test_service_plugins_values(self):
        exp_default_plugins = ['router']
        ovn_rtr_full = 'neutron.services.ovn_l3.plugin.OVNL3RouterPlugin'
        supported_plugins = ['router', 'ovn-router', ovn_rtr_full]
        same_as_input = 'same_as_input'
        TC = namedtuple('TC', 'input exp_plugins exp_uses_rpc description')
        test_cases = [
            TC([], exp_default_plugins, True, "default from empty cfg"),
            TC(['foo'], exp_default_plugins, True,
               "default from unexpected cfg"),
            TC(['foo', 123], exp_default_plugins, True,
               "default from unexpected cfg"),
            TC(['foo', 'router'], exp_default_plugins, True,
               "default from valid cfg"),
            TC(['router'], same_as_input, True, "valid cfg 1"),
            TC(['ovn-router'], same_as_input, False, "valid cfg 2"),
            TC([ovn_rtr_full], same_as_input, False, "valid cfg 3"),
            TC(['ovn-router', ovn_rtr_full, 'router'], supported_plugins, True,
               "valid cfg 4"),
            TC(['router', 'ovn-router'], same_as_input, True,
               "valid cfg 5"),
            TC(['router', ovn_rtr_full], same_as_input, True,
               "valid cfg 6"),
            TC([ovn_rtr_full, 'ovn-router'], ['ovn-router', ovn_rtr_full],
               False, "valid cfg 7"),
            TC(['bar', 'router', 'foo'], ['router'], True, "valid cfg 8"),
            TC(['bar', 'ovn-router', 'foo'], ['ovn-router'], False,
               "valid cfg 9"),
            TC(['bar', ovn_rtr_full, 'foo'], [ovn_rtr_full], False,
               "valid cfg 10"),
            TC(['bar', 'router', 123, 'ovn-router', 'foo', 'kitchen', 'sink',
                ovn_rtr_full], supported_plugins, True, "valid cfg 11"),
        ]
        for tc in test_cases:
            cfg.CONF.set_override("service_plugins", tc.input)
            plugins, rpc_required = pf_plugin._required_service_plugins()
            if tc.exp_plugins == same_as_input:
                self.assertEqual(
                    (tc.input, tc.exp_uses_rpc), (plugins, rpc_required),
                    tc.description)
            else:
                self.assertEqual(
                    (tc.exp_plugins, tc.exp_uses_rpc), (plugins, rpc_required),
                    tc.description)

    @mock.patch.object(cfg.ConfigOpts, '__getattr__')
    def test_service_plugins_no_such_opt(self, mock_config_opts_get):
        description = "test cfg.NoSuchOptError exception"
        mock_config_opts_get.side_effect = cfg.NoSuchOptError('test_svc_plug')
        plugins, rpc_required = pf_plugin._required_service_plugins()
        mock_config_opts_get.assert_called_once()
        expected = (['router'], True)
        self.assertEqual(
            expected, (plugins, rpc_required), description)



# Copyright (c) 2018 Fujitsu Limited
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.agent import l3_extension
from oslo_log import log as logging

from neutron.services.logapi.agent.l3 import base
from neutron.services.logapi.agent import log_extension as log_ext
from neutron.services.logapi.rpc import agent as agent_rpc

LOG = logging.getLogger(__name__)
SNAT_LOG_DRIVER = 'snat_log'


class SNATLoggingExtension(base.L3LoggingExtensionBase,
                           l3_extension.L3AgentExtension):

    def initialize(self, connection, driver_type):
        """Initialize SNAT logging agent extension"""

        snat_log_cls = self._load_driver_cls(
            log_ext.LOGGING_DRIVERS_NAMESPACE, SNAT_LOG_DRIVER)
        self.log_driver = snat_log_cls(self.agent_api)
        self.resource_rpc = agent_rpc.LoggingApiStub()
        self._register_rpc_consumers()
        self.log_driver.initialize(self.resource_rpc)
        LOG.info("Loaded SNAT logging driver")



# Copyright (c) 2016 IBM
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from oslo_config import cfg
from oslo_log import log

from neutron import manager

LOG = log.getLogger(__name__)


class ExternalDNSService(object, metaclass=abc.ABCMeta):
    """Interface definition for an external dns service driver."""

    def __init__(self):
        """Initialize external dns service driver."""

    @classmethod
    def get_instance(cls):
        """Return an instance of the configured external DNS driver."""
        external_dns_driver_name = cfg.CONF.external_dns_driver
        mgr = manager.NeutronManager
        LOG.debug("Loading external dns driver: %s", external_dns_driver_name)
        driver_class = mgr.load_class_for_provider(
            'neutron.services.external_dns_drivers', external_dns_driver_name)
        return driver_class()

    @abc.abstractmethod
    def create_record_set(self, context, dns_domain, dns_name, records):
        """Create a record set in the specified zone.

        :param context: neutron api request context
        :type context: neutron_lib.context.Context
        :param dns_domain: the dns_domain where the record set will be created
        :type dns_domain: String
        :param dns_name: the name associated with the record set
        :type dns_name: String
        :param records: the records in the set
        :type records: List of Strings
        :raises: neutron.extensions.dns.DNSDomainNotFound
                 neutron.extensions.dns.DuplicateRecordSet
        """

    @abc.abstractmethod
    def delete_record_set(self, context, dns_domain, dns_name, records):
        """Delete a record set in the specified zone.

        :param context: neutron api request context
        :type context: neutron.context.Context
        :param dns_domain: the dns_domain from which the record set will be
         deleted
        :type dns_domain: String
        :param dns_name: the dns_name associated with the record set to be
         deleted
        :type dns_name: String
        :param records: the records in the set to be deleted
        :type records: List of Strings
        """



# Copyright 2014 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_service import service

from neutron import service as neutron_service
from neutron.tests.functional import base
from neutron.tests.functional import test_server


class TestService(base.BaseLoggingTestCase):

    def test_api_workers_default(self):
        # This value may end being scaled downward based on available RAM.
        self.assertGreaterEqual(processutils.get_worker_count(),
                                neutron_service._get_api_workers())

    def test_api_workers_from_config(self):
        cfg.CONF.set_override('api_workers', 1234)
        self.assertEqual(1234,
                         neutron_service._get_api_workers())


class TestServiceRestart(test_server.TestNeutronServer):

    def _start_service(self, host, binary, topic, manager, workers,
                       *args, **kwargs):
        server = neutron_service.Service(host, binary, topic, manager,
                                         *args, **kwargs)
        service.launch(cfg.CONF, server, workers).wait()



# Copyright (c) 2019 Intel Corporation.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib import constants
from neutron_lib import context
from neutron_lib import exceptions as exc
from neutron_lib.utils import helpers
from oslo_config import cfg

from neutron.db import segments_db
from neutron.services.network_segment_range import plugin as range_plugin
from neutron.tests.unit.db import test_db_base_plugin_v2 as test_plugin
from neutron.tests.unit import testlib_api

SERVICE_PLUGIN_KLASS = ('neutron.services.network_segment_range.plugin.'
                        'NetworkSegmentRangePlugin')


class TestNetworkSegmentRange(testlib_api.SqlTestCase):

    _foo_range = {'name': 'foo-range',
                  'default': False,
                  'shared': False,
                  'project_id': test_plugin.TEST_TENANT_ID,
                  'network_type': 'foo_network_type',
                  'physical_network': 'foo_phys_net',
                  'minimum': 200,
                  'maximum': 300}

    _flat_range = {'name': 'foo-flat-range',
                   'default': False,
                   'shared': False,
                   'project_id': test_plugin.TEST_TENANT_ID,
                   'network_type': constants.TYPE_FLAT,
                   'physical_network': None,
                   'minimum': 0,
                   'maximum': 0}

    _vlan_range = {'name': 'foo-vlan-range',
                   'default': False,
                   'shared': False,
                   'project_id': test_plugin.TEST_TENANT_ID,
                   'network_type': constants.TYPE_VLAN,
                   'physical_network': 'phys_net',
                   'minimum': 200,
                   'maximum': 300}

    _vxlan_range = {'name': 'foo-vxlan-range',
                    'default': False,
                    'shared': False,
                    'project_id': test_plugin.TEST_TENANT_ID,
                    'network_type': constants.TYPE_VXLAN,
                    'physical_network': None,
                    'minimum': 400,
                    'maximum': 500}

    _gre_range = {'name': 'foo-vlan-range',
                  'default': False,
                  'shared': False,
                  'project_id': test_plugin.TEST_TENANT_ID,
                  'network_type': constants.TYPE_GRE,
                  'physical_network': None,
                  'minimum': 600,
                  'maximum': 700}

    _geneve_range = {'name': 'foo-geneve-range',
                     'default': False,
                     'shared': False,
                     'project_id': test_plugin.TEST_TENANT_ID,
                     'network_type': constants.TYPE_GENEVE,
                     'physical_network': None,
                     'minimum': 800,
                     'maximum': 900}

    def setUp(self):
        super(TestNetworkSegmentRange, self).setUp()
        with mock.patch("neutron_lib.plugins.directory.get_plugin"):
            self.plugin = range_plugin.NetworkSegmentRangePlugin()
        self.context = context.get_admin_context()
        cfg.CONF.set_override('service_plugins', [SERVICE_PLUGIN_KLASS])

    def _validate_resource(self, resource, keys, res_name):
        for k in keys:
            self.assertIn(k, resource[res_name])
            if isinstance(keys[k], list):
                self.assertEqual(
                     sorted(keys[k], key=helpers.safe_sort_key),
                     sorted(resource[res_name][k], key=helpers.safe_sort_key))
            else:
                self.assertEqual(keys[k], resource[res_name][k])

    def test__is_network_segment_range_referenced(self):
        with mock.patch.object(segments_db,
                               'network_segments_exist_in_range',
                               return_value=True):
            self.assertTrue(self.plugin._is_network_segment_range_referenced(
                self.context, self._vlan_range))

    def test__is_network_segment_range_unreferenced(self):
        with mock.patch.object(segments_db,
                               'network_segments_exist_in_range',
                               return_value=False):
            self.assertFalse(self.plugin._is_network_segment_range_referenced(
                self.context, self._vlan_range))

    def test__is_network_segment_range_type_supported(self):
        for foo_range in [self._vlan_range, self._vxlan_range,
                          self._gre_range, self._geneve_range]:
            self.assertTrue(
                self.plugin.
                _is_network_segment_range_type_supported(
                    foo_range['network_type']))

    def test__is_network_segment_range_type_unsupported(self):
        self.assertRaises(
            exc.NeutronException,
            self.plugin._is_network_segment_range_type_supported,
            self._foo_range['network_type'])
        self.assertRaises(
            exc.NeutronException,
            self.plugin._is_network_segment_range_type_supported,
            self._flat_range['network_type'])

    def test__are_allocated_segments_in_range_impacted(self):
        existing_range = self._foo_range
        updated_range = self._vlan_range
        impacted_existing_ranges = [(150, 250), (250, 320),
                                    (200, 300), (180, 330)]
        for ret in impacted_existing_ranges:
            with mock.patch.object(segments_db,
                                   'min_max_actual_segments_in_range',
                                   return_value=ret):
                self.assertTrue(
                    self.plugin._are_allocated_segments_in_range_impacted(
                        self.context, existing_range, updated_range))

    def test__are_allocated_segments_in_range_unimpacted(self):
        existing_range = self._foo_range
        updated_range = self._vlan_range
        with mock.patch.object(segments_db,
                               'min_max_actual_segments_in_range',
                               return_value=(220, 270)):
            self.assertFalse(
                self.plugin._are_allocated_segments_in_range_impacted(
                    self.context, existing_range, updated_range))

    def test_create_network_segment_range(self):
        test_range = self._vlan_range
        network_segment_range = {'network_segment_range': test_range}
        ret = self.plugin.create_network_segment_range(self.context,
                                                       network_segment_range)
        res = {'network_segment_range': ret}
        self._validate_resource(res, test_range, 'network_segment_range')

    def test_create_network_segment_range_failed_with_unsupported_network_type(
            self):
        test_range = self._flat_range
        network_segment_range = {'network_segment_range': test_range}
        self.assertRaises(
            exc.NeutronException,
            self.plugin.create_network_segment_range,
            self.context,
            network_segment_range)

    def test_create_network_segment_range_missing_physical_network_for_vlan(
            self):
        test_range = self._vlan_range.copy()
        test_range.pop("physical_network")
        network_segment_range = {'network_segment_range': test_range}
        self.assertRaises(
            exc.NeutronException,
            self.plugin.create_network_segment_range,
            self.context,
            network_segment_range)

    def test_update_network_segment_range(self):
        test_range = self._vlan_range
        network_segment_range = {'network_segment_range': test_range}
        ret = self.plugin.create_network_segment_range(self.context,
                                                       network_segment_range)
        updated_network_segment_range = {
            'network_segment_range': {'minimum': 700, 'maximum': 800}}
        with mock.patch.object(self.plugin,
                               '_are_allocated_segments_in_range_impacted',
                               return_value=False):
            updated_ret = self.plugin.update_network_segment_range(
                self.context, ret['id'], updated_network_segment_range)
        res = {'network_segment_range': updated_ret}
        test_range['minimum'] = 700
        test_range['maximum'] = 800
        self._validate_resource(res, test_range, 'network_segment_range')

    def test_update_network_segment_range_failed_with_impacted_existing_range(
            self):
        test_range = self._vlan_range
        network_segment_range = {'network_segment_range': test_range}
        ret = self.plugin.create_network_segment_range(self.context,
                                                       network_segment_range)
        updated_network_segment_range = {
            'network_segment_range': {'minimum': 150, 'maximum': 250}}
        with mock.patch.object(self.plugin,
                               '_are_allocated_segments_in_range_impacted',
                               return_value=True):
            self.assertRaises(
                exc.NeutronException,
                self.plugin.update_network_segment_range,
                self.context,
                ret['id'],
                updated_network_segment_range)

    def test_delete_network_segment_range(self):
        test_range = self._vlan_range
        network_segment_range = {'network_segment_range': test_range}
        ret = self.plugin.create_network_segment_range(self.context,
                                                       network_segment_range)
        with mock.patch.object(self.plugin,
                               '_is_network_segment_range_referenced',
                               return_value=False):
            try:
                self.plugin.delete_network_segment_range(
                    self.context, ret['id'])
            except exc.NeutronException:
                self.fail("delete_network_segment_range raised "
                          "NeutronException unexpectedly!")

    def test_delete_network_segment_range_failed_with_segment_referenced(self):
        test_range = self._vlan_range
        network_segment_range = {'network_segment_range': test_range}
        ret = self.plugin.create_network_segment_range(self.context,
                                                       network_segment_range)
        with mock.patch.object(self.plugin,
                               '_is_network_segment_range_referenced',
                               return_value=True):
            self.assertRaises(
                exc.NeutronException,
                self.plugin.delete_network_segment_range,
                self.context,
                ret['id'])
